import mock
import unittest

from xlcalculator.xlfunctions import (
    xl,
    xlerrors,
    func_xltypes,
    date,
    financial
)


class FinancialModuleTest(unittest.TestCase):

    def test_IRR(self):
        self.assertAlmostEqual(
            financial.IRR(
                func_xltypes.Array([[-100, 39, 59, 55, 20]])
            ),
            0.2809484
        )

    def test_IRR_with_guess_non_null(self):
        self.assertAlmostEqual(
            financial.IRR(func_xltypes.Array([[-100, 39, 59, 55, 20]]), 1),
            0.2809484)

    def test_NPV(self):
        self.assertAlmostEqual(
            financial.NPV(0.06, [[1, 2, 3]]), 5.2422470)
        self.assertAlmostEqual(
            financial.NPV(0.06, 1, 2, 3), 5.2422470)
        self.assertAlmostEqual(
            financial.NPV(0.06, 1), 0.9433962)
        self.assertAlmostEqual(
            financial.NPV(0.1, -10000, 3000, 4200, 6800), 1188.44, 2)

        range1 = func_xltypes.Array([[8000, 9200, 10000, 12000, 14500]])
        self.assertAlmostEqual(
            financial.NPV(0.08, range1) + -40000, 1922.06, 2)
        self.assertAlmostEqual(
            financial.NPV(0.08, range1, -9000) + -40000, -3749.47, 2)

    @mock.patch.object(xl, 'COMPATIBILITY', 'PYTHON')
    def test_NPV_in_python_mode(self):
        self.assertAlmostEqual(financial.NPV(0.06, 1, 2, 3), 5.556781772872908)
        self.assertAlmostEqual(financial.NPV(0.06, 1), 1.0)

    def test_NPV_without_values(self):
        self.assertIsInstance(financial.NPV(3.4), xlerrors.ValueExcelError)

    def test_PMT(self):
        self.assertAlmostEqual(
            financial.PMT(0.08 / 12, 10, 10000), -1037.03, 2
        )

    @mock.patch.object(xl, 'COMPATIBILITY', 'PYTHON')
    def test_PMT_in_python_mode(self):
        self.assertAlmostEqual(
            financial.PMT(0.08 / 12, 10, 10000), -1037.03, 2)
        self.assertAlmostEqual(
            financial.PMT(0.08 / 12, 10, 10000, type=1), -1030.164, 2)

    def test_PV(self):
        self.assertAlmostEqual(
            financial.PV(0.08 / 12, 12 * 20, 500.00, 0, 0),
            -59777.15,
            places=2
        )

    def test_SLN(self):
        self.assertEqual(financial.SLN(30000, 7500, 10), 2250)

    def test_VDB(self):
        self.assertAlmostEqual(
            financial.VDB(
                cost=575000,
                salvage=5000,
                life=10,
                start_period=3,
                end_period=5,
                factor=1.5,
            ),
            102160.546875
        )

    def test_VDB_partial(self):
        self.assertAlmostEqual(
            financial.VDB(
                cost=1,
                salvage=0,
                life=14,
                start_period=11.5,
                end_period=12.5,
                factor=1.25,
            ),
            0.068726290454684
        )

    def test_VDB_partial_no_switch(self):
        self.assertAlmostEqual(
            financial.VDB(
                cost=1,
                salvage=0,
                life=5.0,
                start_period=0.5,
                end_period=1.5,
                factor=2.5,
            ),
            0.375
        )

    def test_VDB_non_integer_life(self):
        self.assertAlmostEqual(
            financial.VDB(
                cost=575000,
                salvage=5000,
                life=10.5,
                start_period=3,
                end_period=5,
                factor=1.5,
            ),
            98708.55092429437
        )

    def test_XIRR(self):
        range_00 = func_xltypes.Array(
            [[-10000, 2750, 4250, 3250, 2750]])
        range_01 = func_xltypes.Array([[
            date.DATE(2008, 1, 1),
            date.DATE(2008, 3, 1),
            date.DATE(2008, 10, 30),
            date.DATE(2009, 2, 15),
            date.DATE(2009, 4, 1)
        ]])

        self.assertAlmostEqual(
            financial.XIRR(range_00, range_01, 0.1), 0.373362535)

    def test_XIRR_length_mismatch(self):
        range_00 = [[-10000, 2750, 4250]]
        range_01 = [[date.DATE(2008, 1, 1), date.DATE(2008, 3, 1)]]
        self.assertIsInstance(
            financial.XIRR(range_00, range_01, 0.1), xlerrors.NumExcelError)

    def test_XIRR_not_converge(self):
        dates = [date.DATE(2020, 8, 31),
                 date.DATE(2020, 5, 5),
                 date.DATE(2020, 2, 28),
                 date.DATE(2020, 8, 31),
                 date.DATE(2018, 6, 30)]
        values = [50289.0, -75000.0, 0.0, 0.0, 0.0]
        self.assertIsInstance(
            financial.XIRR(values, dates, 0), xlerrors.NumExcelError)

    def test_XNPV(self):
        range_00 = func_xltypes.Array(
            [[-10000, 2750, 4250, 3250, 2750]])
        range_01 = func_xltypes.Array([[
            date.DATE(2008, 1, 1),
            date.DATE(2008, 3, 1),
            date.DATE(2008, 10, 30),
            date.DATE(2009, 2, 15),
            date.DATE(2009, 4, 1)
        ]])
        self.assertAlmostEqual(
            financial.XNPV(0.09, range_00, range_01), 2086.65, 2)

    def test_XNPV_length_mismatch(self):
        range_00 = [-10000, 2750, 4250]
        range_01 = [date.DATE(2008, 1, 1), date.DATE(2008, 3, 1)]
        self.assertIsInstance(
            financial.XNPV(0.09, range_00, range_01), xlerrors.NumExcelError)

    def test_XNPV_rate_lt_minus_one(self):
        range_00 = func_xltypes.Array(
            [[-10000, 2750, 4250, 3250, 2750]])
        range_01 = func_xltypes.Array([[
            date.DATE(2008, 1, 1),
            date.DATE(2008, 3, 1),
            date.DATE(2008, 10, 30),
            date.DATE(2009, 2, 15),
            date.DATE(2009, 4, 1)
        ]])
        self.assertEqual(
            financial.XNPV(-1.1, range_00, range_01), float('inf'))



from .. import testing


class MatchTest(testing.FunctionalTestCase):
    filename = "MATCH.xlsx"

    def test_evaluation_A8(self):
        # Match defaults.
        excel_value = self.evaluator.get_cell_value('Sheet1!A8')
        value = self.evaluator.evaluate('Sheet1!A8')
        self.assertEqual(excel_value, value)

    def test_evaluation_A9(self):
        # Match defaults.
        excel_value = self.evaluator.get_cell_value('Sheet1!A9')
        value = self.evaluator.evaluate('Sheet1!A9')
        self.assertEqual(excel_value, value)

    def test_evaluation_A10(self):
        # Match defaults.
        excel_value = self.evaluator.get_cell_value('Sheet1!A10')
        value = self.evaluator.evaluate('Sheet1!A10')
        self.assertEqual(excel_value, value)



from .model import ModelCompiler, Model  # noqa
from .evaluator import Evaluator  # noqa


from .xlfunctions.xl import FUNCTIONS, register  # noqa: F401
from .xlfunctions.xlerrors import *  # noqa: F401, F403
from .xlfunctions.func_xltypes import *  # noqa: F401, F403

# Make sure to register all functions
from .xlfunctions import (  # noqa: F401
    date,
    financial,
    information,
    logical,
    lookup,
    math,
    operator,
    statistics,
    text
)


name = "xlcalculator"



import unittest

from xlcalculator.xlfunctions import xlcriteria


class XlCriteriaModuleTest(unittest.TestCase):

    def test_parse_criteria(self):
        check = xlcriteria.parse_criteria('>3')
        self.assertTrue(check(4))
        self.assertFalse(check(2))

    def test_parse_criteria_asBool(self):
        check = xlcriteria.parse_criteria('>3')
        self.assertTrue(check(4))
        self.assertFalse(check(2))

    def test_parse_criteria_with_implicit_operator(self):
        # Assumes equality.
        check = xlcriteria.parse_criteria('1')
        self.assertTrue(check(1))
        self.assertFalse(check(2))

    def test_parse_criteria_with_implicit_operator_string_value(self):
        # Assumes equality.
        check = xlcriteria.parse_criteria('data')
        self.assertTrue(check('data'))
        self.assertFalse(check(2))

    def test_parse_criteria_with_simple_number(self):
        check = xlcriteria.parse_criteria(1)
        self.assertTrue(check(1))
        self.assertFalse(check(2))



from .. import testing


class DateTest(testing.FunctionalTestCase):
    filename = "DATE.xlsx"

    def test_counta_evaluation_A2(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A2')
        value = self.evaluator.evaluate('Sheet1!A2')
        self.assertEqual(excel_value, value)

    def test_evaluation_A3(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A3')
        value = self.evaluator.evaluate('Sheet1!A3')
        self.assertEqual(excel_value, value)

    def test_evaluation_A4(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A4')
        value = self.evaluator.evaluate('Sheet1!A4')
        self.assertEqual(excel_value, value)

    def test_evaluation_A5(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A5')
        value = self.evaluator.evaluate('Sheet1!A5')
        self.assertEqual(excel_value, value)

    def test_evaluation_A6(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A6')
        value = self.evaluator.evaluate('Sheet1!A6')
        self.assertEqual(excel_value, value)

    def test_evaluation_A7(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A7')
        value = self.evaluator.evaluate('Sheet1!A7')
        self.assertEqual(excel_value, value)

    def test_evaluation_A9(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A9')
        value = self.evaluator.evaluate('Sheet1!A9')
        self.assertEqual(excel_value, value)

    def test_evaluation_A10(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A10')
        value = self.evaluator.evaluate('Sheet1!A10')
        self.assertEqual(excel_value, value)

    def test_evaluation_A11(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A11')
        value = self.evaluator.evaluate('Sheet1!A11')
        self.assertEqual(excel_value, value)

    def test_evaluation_A12(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A12')
        value = self.evaluator.evaluate('Sheet1!A12')
        self.assertEqual(excel_value, value)

    def test_evaluation_A13(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A13')
        value = self.evaluator.evaluate('Sheet1!A13')
        self.assertEqual(excel_value, value)

    def test_evaluation_A14(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A14')
        value = self.evaluator.evaluate('Sheet1!A14')
        self.assertEqual(excel_value, value)

    def test_evaluation_A15(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A15')
        value = self.evaluator.evaluate('Sheet1!A15')
        self.assertEqual(excel_value, value)

    def test_evaluation_A16(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A16')
        value = self.evaluator.evaluate('Sheet1!A16')
        self.assertEqual(excel_value, value)

    def test_evaluation_A17(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A17')
        value = self.evaluator.evaluate('Sheet1!A17')
        self.assertEqual(excel_value, value)

    def test_evaluation_A18(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A18')
        value = self.evaluator.evaluate('Sheet1!A18')
        self.assertEqual(excel_value, value)

    def test_evaluation_A19(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A19')
        value = self.evaluator.evaluate('Sheet1!A19')
        self.assertEqual(excel_value, value)

    def test_evaluation_A25(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A25')
        value = self.evaluator.evaluate('Sheet1!A25')
        self.assertEqual(excel_value, value)

    def test_evaluation_B25(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!B25')
        value = self.evaluator.evaluate('Sheet1!B25')
        self.assertEqual(excel_value, value)

    def test_evaluation_C25(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!C25')
        value = self.evaluator.evaluate('Sheet1!C25')
        self.assertEqual(excel_value, value)

    def test_evaluation_D25(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!D25')
        value = self.evaluator.evaluate('Sheet1!D25')
        self.assertEqual(excel_value, value)

    def test_evaluation_E25(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!E25')
        value = self.evaluator.evaluate('Sheet1!E25')
        self.assertEqual(excel_value, value)

    def test_evaluation_F25(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!F25')
        value = self.evaluator.evaluate('Sheet1!F25')
        self.assertEqual(excel_value, value)




ERROR_CODE_NULL = '#NULL!'
ERROR_CODE_DIV_ZERO = "#DIV/0!"
ERROR_CODE_VALUE = "#VALUE!"
ERROR_CODE_REF = "#REF!"
ERROR_CODE_NAME = "#NAME?"
ERROR_CODE_NUM = "#NUM!"
ERROR_CODE_NA = "#N/A"

ERROR_CODES = (
    ERROR_CODE_NULL,
    ERROR_CODE_DIV_ZERO,
    ERROR_CODE_VALUE,
    ERROR_CODE_REF,
    ERROR_CODE_NAME,
    ERROR_CODE_NUM,
    ERROR_CODE_NA,
)

ERRORS_BY_CODE = {}


def register(cls):
    ERRORS_BY_CODE[cls.value] = cls
    return cls


class ExcelError(Exception):

    def __init__(self, value, info=None):
        super().__init__(info)
        self.value = value
        self.info = info

    @classmethod
    def is_error(cls, value):
        return isinstance(value, cls)

    def __str__(self):
        return str(self.value)

    def __eq__(self, other):
        if isinstance(other, str):
            return str(self) == other
        return id(self) == id(other)


class SpecificExcelError(ExcelError):
    value = None

    def __init__(self, info=None):
        super().__init__(self.value, info)


@register
class NullExcelError(SpecificExcelError):
    value = ERROR_CODE_NULL


@register
class DivZeroExcelError(SpecificExcelError):
    value = ERROR_CODE_DIV_ZERO


@register
class ValueExcelError(SpecificExcelError):
    value = ERROR_CODE_VALUE


@register
class RefExcelError(SpecificExcelError):
    value = ERROR_CODE_REF


@register
class NameExcelError(SpecificExcelError):
    value = ERROR_CODE_NAME


@register
class NumExcelError(SpecificExcelError):
    value = ERROR_CODE_NUM


@register
class NaExcelError(SpecificExcelError):
    value = ERROR_CODE_NA



import datetime

EXCEL_EPOCH = datetime.datetime(1900, 1, 1)


def number_to_datetime(value):
    offset = 2 if value > 58 else 1
    delta = datetime.timedelta(
        days=int(value) - offset, seconds=(value % 1) * 24 * 60 * 60)
    return EXCEL_EPOCH + delta


def datetime_to_number(value):
    delta = value - EXCEL_EPOCH
    # Excel treats 1900 as a leap year.
    offset = 2 if delta.days > 58 else 1
    return (delta.days + offset) + (delta.seconds / 24 * 60 * 60)



from .. import testing


class ABSTest(testing.FunctionalTestCase):

    filename = "ABS.xlsx"

    def test_evaluation_A1(self):
        self.assertEqual(
            self.evaluator.evaluate('Sheet1!A1'),
            self.evaluator.get_cell_value('Sheet1!A1')
        )



import mock
import unittest

from xlcalculator.xlfunctions import xlerrors, func_xltypes
from xlcalculator import ast_nodes, xltypes
from xlcalculator.tokenizer import f_token


def context(ref, model=None):
    ctx = ast_nodes.EvalContext(ref=ref)
    if model is not None:
        ctx.eval_cell = lambda addr: model.cells[addr].value
        ctx.cells = model.cells
        ctx.ranges = model.ranges
    return ctx


class EvalContextTest(unittest.TestCase):

    def test__init__(self):
        ctx = ast_nodes.EvalContext(ref='A1')
        self.assertEqual(ctx.seen, [])
        self.assertIsInstance(ctx.namespace, dict)
        self.assertEqual(ctx.ref, 'A1')

    def test_eval_cell(self):
        ctx = ast_nodes.EvalContext(ref='A1')
        with self.assertRaises(NotImplementedError):
            ctx.eval_cell('A2')


class ASTNodeTest(unittest.TestCase):

    def create_node(self):
        return ast_nodes.ASTNode(
            f_token(tvalue='tv', ttype='tt', tsubtype='tst'))

    def test_init(self):
        node = self.create_node()
        self.assertIsNotNone(node.token)

    def test_getattr(self):
        node = self.create_node()
        self.assertEqual(node.tvalue, 'tv')
        self.assertEqual(node.ttype, 'tt')
        self.assertEqual(node.tsubtype, 'tst')

    def test_eq(self):
        node1 = self.create_node()
        node2 = self.create_node()
        self.assertEqual(node1, node1)
        # Even though the two nodes have the same token values, each token has
        # a UUID that is used for comparison.
        self.assertNotEqual(node1, node2)

    def test_eval(self):
        node = self.create_node()
        with self.assertRaises(NotImplementedError):
            node.eval(context('A1'))

    def test_repr(self):
        node = self.create_node()
        self.assertEqual(
            repr(node), "<ASTNode tvalue: 'tv', ttype: tt, tsubtype: tst>")

    def test_str(self):
        node = self.create_node()
        self.assertEqual(str(node), 'tv')

    def test_iter(self):
        node = self.create_node()
        self.assertEqual(tuple(node), (node,))


class OperandNodeTest(unittest.TestCase):

    def create_node(self, value='1', type='nuber'):
        return ast_nodes.OperandNode(
            f_token(tvalue=value, ttype='operand', tsubtype=type))

    def test_eval(self):
        node = self.create_node()
        self.assertEqual(node.eval(context('A1')), 1)

    def test_eval_bool(self):
        node = self.create_node('TRUE', 'logical')
        self.assertEqual(node.eval(context('A1')), True)

    def test_eval_text(self):
        node = self.create_node('data', 'text')
        self.assertEqual(node.eval(context('A1')), 'data')

    def test_eval_error(self):
        node = self.create_node('#VALUE!', 'error')
        self.assertIsInstance(
            node.eval(context('A1')), xlerrors.ValueExcelError)

    def test_eval_error_with_nonstandard_error_value(self):
        node = self.create_node('ERR:504', 'error')
        self.assertIsInstance(
            node.eval(context('A1')), xlerrors.ExcelError)

    def test_str(self):
        node = self.create_node()
        self.assertEqual(str(node), '1')

    def test_str_bool(self):
        node = self.create_node('TRUE', 'logical')
        self.assertEqual(str(node), 'True')

    def test_str_text(self):
        node = self.create_node('data', 'text')
        self.assertEqual(str(node), '"data"')
        node = self.create_node('This is "data"', 'text')
        self.assertEqual(str(node), '"This is \\"data\\""')


class RangeNodeTest(unittest.TestCase):

    def setUp(self):
        self.model = mock.Mock(
            cells={
                'Sh1!' + addr: xltypes.XLCell('Sh1!' + addr, value=idx)
                for idx, addr in enumerate(('A1', 'A2', 'B1', 'B2'))
            },
            ranges={'Sh1!A1:B2': xltypes.XLRange('Sh1!A1:B2')},
            defined_names={},
        )
        self.model.defined_names['first'] = self.model.ranges['Sh1!A1:B2']

    def create_node(self, range):
        return ast_nodes.RangeNode(
            f_token(tvalue=range, ttype='operand', tsubtype='range'))

    def test_get_cells(self):
        self.assertEqual(
            self.create_node('A1').get_cells(), ['A1'])
        self.assertEqual(
            self.create_node('A1:A3').get_cells(), [['A1'], ['A2'], ['A3']])
        self.assertEqual(
            self.create_node('S!A1:A2').get_cells(), [['S!A1'], ['S!A2']])

    def test_eval(self):
        node = self.create_node('A1:B2')
        res = node.eval(context('Sh1!C1', self.model))
        exp = func_xltypes.Array([[0, 2], [1, 3]])
        self.assertTrue((res == exp).all().all())

    def test_eval_cell(self):
        node = self.create_node('A1')
        self.assertEqual(node.eval(context('Sh1!C1', self.model)), 0)

    def test_str(self):
        node = self.create_node('A1:B2')
        self.assertEqual(str(node), 'A1:B2')


class OperatorNodeTest(unittest.TestCase):

    def create_node(self):
        node = ast_nodes.OperatorNode(
            f_token(tvalue='+', ttype='operator-infix', tsubtype='math'))
        node.left = ast_nodes.OperandNode(
            f_token(tvalue='1', ttype='operand', tsubtype='number'))
        node.right = ast_nodes.OperandNode(
            f_token(tvalue='2', ttype='operand', tsubtype='number'))
        return node

    def test_eval(self):
        node = self.create_node()
        self.assertEqual(node.eval(context('A1')), 3)

    def test_eval_prefix(self):
        node = ast_nodes.OperatorNode(
            f_token(tvalue='-', ttype='operator-prefix', tsubtype='math'))
        node.right = ast_nodes.OperandNode(
            f_token(tvalue='1', ttype='operand', tsubtype='number'))
        self.assertEqual(node.eval(context('A1')), -1)

    def test_eval_postfix(self):
        node = ast_nodes.OperatorNode(
            f_token(tvalue='%', ttype='operator-postfix', tsubtype='math'))
        node.left = ast_nodes.OperandNode(
            f_token(tvalue='1', ttype='operand', tsubtype='number'))
        self.assertEqual(node.eval(context('A1')), 0.01)

    def test_eval_unknown_type(self):
        node = ast_nodes.OperatorNode(
            f_token(tvalue='-', ttype='operator', tsubtype='math'))
        with self.assertRaises(ValueError):
            node.eval(context('A1'))

    def test_str(self):
        node = self.create_node()
        self.assertEqual(str(node), '(1) + (2)')

    def test_str_prefix(self):
        node = ast_nodes.OperatorNode(
            f_token(tvalue='-', ttype='operator-prefix', tsubtype='math'))
        node.right = ast_nodes.OperandNode(
            f_token(tvalue='1', ttype='operand', tsubtype='number'))
        self.assertEqual(str(node), '- (1)')

    def test_str_postfix(self):
        node = ast_nodes.OperatorNode(
            f_token(tvalue='%', ttype='operator-postfix', tsubtype='math'))
        node.left = ast_nodes.OperandNode(
            f_token(tvalue='1', ttype='operand', tsubtype='number'))
        self.assertEqual(str(node), '(1) %')

    def test_iter(self):
        node = self.create_node()
        self.assertEqual(tuple(node), (node.left, node.right, node))


class FunctionNodeTest(unittest.TestCase):

    def create_node(self):
        node = ast_nodes.FunctionNode(
            f_token(tvalue='MOD', ttype='function', tsubtype='start'))
        node.args = [
            ast_nodes.OperandNode(
                f_token(tvalue='3', ttype='operand', tsubtype='number')),
            ast_nodes.OperandNode(
                f_token(tvalue='2', ttype='operand', tsubtype='number')),
        ]
        return node

    def test_eval(self):
        node = self.create_node()
        self.assertEqual(node.eval(context('A1')), 1)

    def test_eval_var_positional(self):
        node = ast_nodes.FunctionNode(
            f_token(tvalue='SUM', ttype='function', tsubtype='start'))
        node.args = [
            ast_nodes.OperandNode(
                f_token(tvalue='3', ttype='operand', tsubtype='number')),
            ast_nodes.OperandNode(
                f_token(tvalue='2', ttype='operand', tsubtype='number')),
        ]
        self.assertEqual(node.eval(context('A1')), 5)

    def test_eval_expr(self):
        cond_node = ast_nodes.OperatorNode(
            f_token(tvalue='>', ttype='operator-infix', tsubtype='math')
        )
        cond_node.left = ast_nodes.OperandNode(
            f_token(tvalue='1', ttype='operand', tsubtype='number'))
        cond_node.right = ast_nodes.OperandNode(
            f_token(tvalue='0', ttype='operand', tsubtype='number'))

        node = ast_nodes.FunctionNode(
            f_token(tvalue='IF', ttype='function', tsubtype='start'))
        node.args = [
            cond_node,
            ast_nodes.OperandNode(
                f_token(tvalue='3', ttype='operand', tsubtype='number')),
            ast_nodes.OperandNode(
                f_token(tvalue='2', ttype='operand', tsubtype='number')),
        ]
        self.assertEqual(node.eval(context('A1')), 3)

    def test_eval_expr_var_positional(self):
        node = ast_nodes.FunctionNode(
            f_token(tvalue='AND', ttype='function', tsubtype='start'))
        node.args = [
            ast_nodes.OperandNode(
                f_token(tvalue='3', ttype='operand', tsubtype='number')),
            ast_nodes.OperandNode(
                f_token(tvalue='2', ttype='operand', tsubtype='number')),
        ]
        self.assertEqual(node.eval(context('A1')), True)

    def test_str(self):
        node = self.create_node()
        self.assertEqual(str(node), 'MOD(3, 2)')

    def test_iter(self):
        node = self.create_node()
        self.assertEqual(tuple(node), (node.args[0], node.args[1], node))




import unittest
from copy import deepcopy

from jsonpickle import decode

from xlcalculator.model import Model, ModelCompiler
from xlcalculator.xltypes import XLCell, XLFormula, XLRange
from xlcalculator.tokenizer import f_token
from xlcalculator import Evaluator

from . import testing


class ModelTest(testing.XlCalculatorTestCase):

    def setUp(self):
        infile = open(testing.get_resource("model.json"), "rb")
        json_bytes = infile.read()
        infile.close()
        data = decode(
            json_bytes, keys=True,
            classes=(XLCell, XLFormula, f_token, XLRange))
        self.cells = data['cells']
        self.defined_names = data['defined_names']
        self.ranges = data['ranges']
        self.formulae = data['formulae']

        self.model = Model()
        self.model.cells = deepcopy(self.cells)
        self.model.defined_names = deepcopy(self.defined_names)
        self.model.ranges = deepcopy(self.ranges)
        self.model.formulae = deepcopy(self.formulae)

    def test_build_code(self):

        this_model = Model()
        this_model.cells = self.cells
        this_model.defined_names = self.defined_names
        this_model.ranges = self.ranges
        this_model.formulae = self.formulae

        this_model.build_code()

        self.assertEqual(self.model, this_model)

    def test_set_value(self):
        this_model = deepcopy(self.model)

        # Cells we know are in the model.

        this_model.set_cell_value('First!A2', 88)
        cell_value_00 = 88
        self.assertEqual(cell_value_00, this_model.cells['First!A2'].value)

        this_model.set_cell_value(XLCell('First!A2', 22), 0.1)
        cell_value_01 = 0.1
        self.assertEqual(cell_value_01, this_model.cells['First!A2'].value)

        this_model.set_cell_value('First!A1', 'Header_1')
        cell_value_00 = 'Header_1'
        self.assertEqual(cell_value_00, this_model.cells['First!A1'].value)

        # Cells we know are not in the model, they get added

        this_model.set_cell_value('Sheet1!A1', 88)
        cell_value_00 = 88
        self.assertEqual(cell_value_00, this_model.cells['Sheet1!A1'].value)

        this_model.set_cell_value(XLCell('Sheet1!A1', 22), 0.1)
        cell_value_01 = 0.1
        self.assertEqual(cell_value_01, this_model.cells['Sheet1!A1'].value)

    def test_get_value(self):
        this_model = deepcopy(self.model)

        # cells we know are in the model
        get_cell_value_00 = this_model.get_cell_value('First!A2')
        cell_value_00 = 0.1
        self.assertEqual(cell_value_00, get_cell_value_00)

        get_cell_value_01 = this_model.get_cell_value('First!A1')
        cell_value_01 = 'Header 1'
        self.assertEqual(cell_value_01, get_cell_value_01)

        get_cell_value_02 = this_model.get_cell_value(XLCell('First!A1', 22))
        cell_value_02 = 'Header 1'
        self.assertEqual(cell_value_02, get_cell_value_02)

        get_cell_value_03 = this_model.get_cell_value(XLCell('First!A2', 22))
        cell_value_03 = 0.1
        self.assertEqual(cell_value_03, get_cell_value_03)

        # Cells we know are not in the model, all default to the value 0

        get_cell_value_04 = this_model.get_cell_value('Sheet1!A1')
        cell_value_04 = 0
        self.assertEqual(cell_value_04, get_cell_value_04)


class ModelCompilerTest(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        infile = open(testing.get_resource("reader.json"), "rb")
        json_bytes = infile.read()
        infile.close()
        data = decode(
            json_bytes, keys=True,
            classes=(XLCell, XLFormula, f_token, XLRange))
        self.cells = data['cells']
        self.defined_names = data['defined_names']
        self.ranges = data['ranges']
        self.formulae = data['formulae']

        self.model = Model()
        self.model.cells = deepcopy(self.cells)
        self.model.defined_names = deepcopy(self.defined_names)
        self.model.ranges = deepcopy(self.ranges)
        self.model.formulae = deepcopy(self.formulae)

    def test_read_and_parse_archive(self):
        model_compiler = ModelCompiler()
        new_model = model_compiler.read_and_parse_archive(
            testing.get_resource("reader.xlsm"), ignore_sheets=['Eleventh'])
        self.assertEqual(self.model, new_model)

    def test_build_defined_names(self):
        model_compiler = ModelCompiler()
        archive = model_compiler.read_excel_file(
            testing.get_resource("reader.xlsm"))
        (
            model_compiler.model.cells,
            model_compiler.model.formulae,
            model_compiler.model.ranges
        ) = archive.read_cells(ignore_sheets=['Eleventh'])
        model_compiler.defined_names = archive.read_defined_names()
        model_compiler.build_defined_names()

        self.assertEqual(
            self.model.defined_names,
            model_compiler.model.defined_names
        )

    def test_link_cells_to_defined_names(self):
        model_compiler = ModelCompiler()
        archive = model_compiler.read_excel_file(
            testing.get_resource("reader.xlsm"))
        (
            model_compiler.model.cells,
            model_compiler.model.formulae,
            model_compiler.model.ranges
        ) = archive.read_cells(ignore_sheets=['Eleventh'])
        model_compiler.defined_names = archive.read_defined_names()
        model_compiler.build_defined_names()
        model_compiler.link_cells_to_defined_names()

        self.assertEqual(self.model.cells, model_compiler.model.cells)

    @unittest.skip(
        'The ranges do not line up anymore until reader.json get regenerated. '
        'Shared formulas do not produce ranges anymore as they are resolved '
        'by the reader automatically.'
    )
    def test_build_ranges(self):
        model_compiler = ModelCompiler()
        archive = model_compiler.read_excel_file(
            testing.get_resource("reader.xlsm"))
        (
            model_compiler.model.cells,
            model_compiler.model.formulae,
            model_compiler.model.ranges
        ) = archive.read_cells(ignore_sheets=['Eleventh'])
        model_compiler.defined_names = archive.read_defined_names()
        model_compiler.build_defined_names()
        model_compiler.build_ranges()

        self.assertEqual(self.model.ranges, model_compiler.model.ranges)

    def test_extract_cells(self):
        model_compiler = ModelCompiler()
        reader_model = model_compiler.read_and_parse_archive(
            testing.get_resource("reader.xlsm"), ignore_sheets=['Eleventh'])
        extracted_model = ModelCompiler.extract(
            reader_model, focus=['First!A2', 'First!B2', 'First!C2'])

        reference_model = Model()
        reference_model.set_cell_value('First!A2', 0.1)
        reference_model.set_cell_value('First!B2', 0.2)
        reference_model.set_cell_value('First!C2', 0.3)

        self.assertEqual(reference_model.cells, extracted_model.cells)

    def test_extract_defined_names(self):
        model_compiler = ModelCompiler()
        reader_model = model_compiler.read_and_parse_archive(
            testing.get_resource("reader.xlsm"), ignore_sheets=['Eleventh'])
        extracted_model = ModelCompiler.extract(
            reader_model, focus=['Hundred', 'My_Range'])

        reference_model = Model()
        reference_model.set_cell_value('Eighth!B1', 100)
        reference_model.set_cell_value('Eighth!A1', 1)
        reference_model.set_cell_value('Eighth!A2', 2)
        reference_model.set_cell_value('Eighth!A3', 3)
        reference_model.set_cell_value('Eighth!A4', 4)
        reference_model.set_cell_value('Eighth!A5', 5)
        reference_model.set_cell_value('Eighth!A6', 6)
        reference_model.set_cell_value('Eighth!A7', 7)
        reference_model.set_cell_value('Eighth!A8', 8)
        reference_model.set_cell_value('Eighth!A9', 9)
        reference_model.set_cell_value('Eighth!A10', 10)
        reference_model.defined_names['Hundred'] = deepcopy(
            reader_model.defined_names['Hundred'])
        reference_model.defined_names['My_Range'] = deepcopy(
            reader_model.defined_names['My_Range'])

        self.assertEqual(reference_model.cells, extracted_model.cells)
        self.assertEqual(
            reference_model.defined_names, extracted_model.defined_names)
        self.assertEqual(reference_model, extracted_model)

    def test_extract(self):
        model_compiler = ModelCompiler()
        reader_model = model_compiler.read_and_parse_archive(
            testing.get_resource("reader.xlsm"), ignore_sheets=['Eleventh'])
        extracted_model = ModelCompiler.extract(
            reader_model,
            focus=['First!A2', 'First!B2', 'First!C2', 'Fourth!A2',
                   'Hundred', 'My_Range']
        )

        reference_model = Model()
        reference_model.set_cell_value('First!A2', 0.1)
        reference_model.set_cell_value('First!B2', 0.2)
        reference_model.set_cell_value('First!C2', 0.3)
        reference_model.set_cell_value('Eighth!B1', 100)
        reference_model.set_cell_value('Eighth!A1', 1)
        reference_model.set_cell_value('Eighth!A2', 2)
        reference_model.set_cell_value('Eighth!A3', 3)
        reference_model.set_cell_value('Eighth!A4', 4)
        reference_model.set_cell_value('Eighth!A5', 5)
        reference_model.set_cell_value('Eighth!A6', 6)
        reference_model.set_cell_value('Eighth!A7', 7)
        reference_model.set_cell_value('Eighth!A8', 8)
        reference_model.set_cell_value('Eighth!A9', 9)
        reference_model.set_cell_value('Eighth!A10', 10)
        reference_model.defined_names['Hundred'] = deepcopy(
            reader_model.defined_names['Hundred'])
        reference_model.defined_names['My_Range'] = deepcopy(
            reader_model.defined_names['My_Range'])
        reference_model.cells['Fourth!A2'] = reader_model.cells['Fourth!A2']

        self.assertEqual(reference_model.cells, extracted_model.cells)
        self.assertEqual(
            reference_model.defined_names,
            extracted_model.defined_names)

    def test_extract_and_evaluate(self):
        model_compiler = ModelCompiler()
        reader_model = model_compiler.read_and_parse_archive(
            testing.get_resource("model_compiler_and_evaluate.xlsx"))
        extracted_model = ModelCompiler.extract(
            reader_model,
            focus=['add_one']
        )

        self.assertEqual(1, extracted_model.get_cell_value('Sheet1!A1'))
        self.assertEqual(2, extracted_model.get_cell_value('Sheet1!B1'))
        extracted_model.set_cell_value('Sheet1!A1', 2)
        evaluator = Evaluator(extracted_model)
        evaluator.evaluate('add_one')
        self.assertEqual(3, extracted_model.get_cell_value('Sheet1!B1'))

    def test_dict_read_and_parse(self):

        input_dict = {
            "B4": 0.95,
            "B2": 1000,
            "B19": 0.001,
            "B20": 4,
            # B21
            "B22": 1,
            "B23": 2,
            "B24": 3,
            "B25": "=B2*B4",
            "B26": 5,
            "B27": 6,
            "B28": "=B19*B20*B22",
            "C22": "=SUM(B22:B28)",
        }

        model_compiler = ModelCompiler()
        my_model = model_compiler.read_and_parse_dict(input_dict)
        evaluator = Evaluator(my_model)

        # cells need a sheet and Sheet1 is default.
        evaluator.set_cell_value("Sheet1!B22", 100)
        self.assertEqual(0.4, evaluator.evaluate("Sheet1!B28"))
        self.assertEqual(1066.4, evaluator.evaluate("Sheet1!C22"))



import openpyxl

from . import patch, xltypes


class Reader():

    def __init__(self, file_name):
        self.excel_file_name = file_name

    def read(self):
        with patch.openpyxl_WorksheetReader_patch():
            self.book = openpyxl.load_workbook(self.excel_file_name)

    def read_defined_names(self, ignore_sheets=[], ignore_hidden=False):
        return {
            defn.name: defn.value
            for defn in self.book.defined_names.definedName
            if defn.hidden is None and defn.value != '#REF!'
        }

    def read_cells(self, ignore_sheets=[], ignore_hidden=False):
        cells = {}
        formulae = {}
        ranges = {}
        for sheet_name in self.book.sheetnames:
            if sheet_name in ignore_sheets:
                continue
            sheet = self.book[sheet_name]
            for cell in sheet._cells.values():
                addr = f'{sheet_name}!{cell.coordinate}'
                if cell.data_type == 'f':
                    formula = xltypes.XLFormula(cell.value, sheet_name)
                    formulae[addr] = formula
                    value = cell.cvalue
                else:
                    formula = None
                    value = cell.value

                cells[addr] = xltypes.XLCell(
                    addr, value=value, formula=formula)

        return [cells, formulae, ranges]



import unittest

from xlcalculator.xlfunctions import xlerrors, operator


class OperatorModuleTest(unittest.TestCase):

    def test_OP_MUL(self):
        self.assertEqual(operator.OP_MUL(1, 2), 2)

    def test_OP_DIV(self):
        self.assertEqual(operator.OP_DIV(2, 1), 2)

    def test_OP_DIV_by_zero(self):
        self.assertIsInstance(
            operator.OP_DIV(2, 0), xlerrors.DivZeroExcelError)

    def test_OP_ADD(self):
        self.assertEqual(operator.OP_ADD(1, 2), 3)

    def test_OP_SUB(self):
        self.assertEqual(operator.OP_SUB(1, 2), -1)

    def test_OP_EQ(self):
        self.assertEqual(operator.OP_EQ(2, 2), True)
        self.assertEqual(operator.OP_EQ("2", 2), False)
        self.assertEqual(operator.OP_EQ(True, 1), True)
        self.assertEqual(operator.OP_EQ(1, 2), False)

    def test_OP_NE(self):
        self.assertEqual(operator.OP_NE(2, 2), False)
        self.assertEqual(operator.OP_NE("2", 2), True)
        self.assertEqual(operator.OP_NE(True, 1), False)
        self.assertEqual(operator.OP_NE(1, 2), True)

    def test_OP_GT(self):
        self.assertEqual(operator.OP_GT(2, 1), True)
        self.assertEqual(operator.OP_GT("2", 1), True)
        self.assertEqual(operator.OP_GT(1, 2), False)

    def test_OP_GT_with_None(self):
        self.assertEqual(operator.OP_GT(None, 2), False)
        self.assertEqual(operator.OP_GT(2, None), False)

    def test_OP_LT(self):
        self.assertEqual(operator.OP_LT(2, 1), False)
        self.assertEqual(operator.OP_LT("2", 1), False)
        self.assertEqual(operator.OP_LT(1, 2), True)

    def test_OP_LT_with_None(self):
        self.assertEqual(operator.OP_LT(None, 2), False)
        self.assertEqual(operator.OP_LT(2, None), False)

    def test_OP_GE(self):
        self.assertEqual(operator.OP_GE(2, 1), True)
        self.assertEqual(operator.OP_GE(1, 1), True)
        self.assertEqual(operator.OP_GE(1, 2), False)

    def test_OP_GE_with_None(self):
        self.assertEqual(operator.OP_GE(None, 2), False)
        self.assertEqual(operator.OP_GE(2, None), False)

    def test_OP_LE(self):
        self.assertEqual(operator.OP_LE(2, 1), False)
        self.assertEqual(operator.OP_LE(1, 1), True)
        self.assertEqual(operator.OP_LE(1, 2), True)

    def test_OP_LE_with_None(self):
        self.assertEqual(operator.OP_LE(None, 2), False)
        self.assertEqual(operator.OP_LE(2, None), False)

    def test_OP_NEG(self):
        self.assertEqual(operator.OP_NEG(1), -1)
        self.assertEqual(operator.OP_NEG(-1), 1)

    def test_OP_PERCENT(self):
        self.assertEqual(operator.OP_PERCENT(2), 0.02)




import logging
from xlcalculator import ModelCompiler
from xlcalculator import Model
from xlcalculator import Evaluator

logging.basicConfig(level=logging.DEBUG)

json_file_name = r'use_case_01.json'

filename = r'use_case_01.xlsm'
compiler = ModelCompiler()
new_model = compiler.read_and_parse_archive(filename, build_code=False)
new_model.persist_to_json_file(json_file_name)

reconstituted_model = Model()
reconstituted_model.construct_from_json_file(json_file_name, build_code=True)

evaluator = Evaluator(reconstituted_model)
val1 = evaluator.evaluate('First!A2')
print("value 'evaluated' for First!A2 without a formula:", val1)
val2 = evaluator.evaluate('Seventh!C1')
print("value 'evaluated' for Seventh!C1 with a formula:", val2)
val3 = evaluator.evaluate('Ninth!B1')
print("value 'evaluated' for Ninth!B1 with a defined name:", val3)
val4 = evaluator.evaluate('Hundred')
print("value 'evaluated' for Hundred with a defined name:", val4)
val5 = evaluator.evaluate('Tenth!C1')
print("value 'evaluated' for Tenth!C1 with a defined name:", val5)
val6 = evaluator.evaluate('Tenth!C2')
print("value 'evaluated' for Tenth!C2 with a defined name:", val6)
val7 = evaluator.evaluate('Tenth!C3')
print("value 'evaluated' for Tenth!C3 with a defined name:", val7)

evaluator.set_cell_value('First!A2', 88)
val17 = evaluator.evaluate('First!A2')
print("New value for First!A2 is", val17)



import sys
from functools import lru_cache

from xlcalculator.xlfunctions import xl, func_xltypes

from . import ast_nodes, xltypes


class EvaluatorContext(ast_nodes.EvalContext):

    def __init__(self, evaluator, ref):
        super().__init__(evaluator.namespace, ref)
        self.evaluator = evaluator

    @property
    def cells(self):
        return self.evaluator.model.cells

    @property
    def ranges(self):
        return self.evaluator.model.ranges

    @lru_cache(maxsize=None)
    def eval_cell(self, addr):
        # Check for a cycle.
        if addr in self.seen:
            raise RuntimeError(
                f'Cycle detected for {addr}:\n- ' + '\n- '.join(self.seen))
        self.seen.append(addr)
        self.set_sheet(sheet=addr.split("!")[0])

        return self.evaluator.evaluate(addr, self)


class Evaluator:
    """Traverses and evaluates a given model."""

    def __init__(self, model, namespace=None):
        self.model = model
        self.namespace = namespace \
            if namespace is not None else xl.FUNCTIONS.copy()
        self.cache_count = 0

    def _get_context(self, ref):
        return EvaluatorContext(self, ref)

    def resolve_names(self, addr):
        # Although defined names have been resolved in Model.create_node()
        # we need to attempt to resolve defined names as we might have been
        # given one in argument addr.
        if addr not in self.model.defined_names:
            return addr

        defn = self.model.defined_names[addr]

        if isinstance(defn, xltypes.XLCell):
            return defn.address

        if isinstance(defn, xltypes.XLRange):
            raise ValueError(
                f"I can't resolve {addr} to a cell. It's a "
                f"range and they aren't supported yet.")

        if isinstance(defn, xltypes.XLFormula):
            raise ValueError(
                f"I can't resolve {addr} to a cell. It's a "
                f"formula and they aren't supported as a cell "
                f"reference.")

    def evaluate(self, addr, context=None):
        # 1. Resolve the address to a cell.
        addr = self.resolve_names(addr)
        if addr not in self.model.cells:
            # Blank cell that has no stored value in the model.
            return func_xltypes.BLANK
        cell = self.model.cells[addr]

        # 2. If there is no formula, we simply return the cell value.
        if (cell.formula is None or cell.formula.evaluate is False):
            return func_xltypes.ExcelType.cast_from_native(
                self.model.cells[addr].value)

        # 3. Prepare the execution environment and evaluate the formula.
        #    (Note: Range nodes will automatically evaluate all their
        #           dependencies.)
        context = context if context is not None else self._get_context(addr)
        try:
            value = cell.formula.ast.eval(context)
        except Exception as err:
            raise RuntimeError(
                f"Problem evaluating cell {addr} formula "
                f"{cell.formula.formula}: {repr(err)}"
            ).with_traceback(sys.exc_info()[2])

        # 4. Update the cell value.
        #    Note for later: If an array is returned, we should distribute the
        #    values to the respective cell (known as spilling).
        cell.value = value
        cell.need_update = False

        return value

    def set_cell_value(self, address, value):
        """Sets the value of a cell in the model."""
        self.model.set_cell_value(address, value)

    def get_cell_value(self, address):
        """Gets the value of a cell in the model."""
        return self.model.get_cell_value(address)



import unittest

from xlcalculator.model import parser


class FormulaParserTest(unittest.TestCase):

    def parse(self, formula, named_ranges=None):
        if named_ranges is None:
            named_ranges = {
                'first': 'A1',
                'two_by_two': 'A1:B2'
            }
        ast = parser.FormulaParser().parse(formula, named_ranges)
        return ast

    def test_parse(self):
        self.assertEqual(str(self.parse('=A1+1')), '(A1) + (1)')

    def test_parse_without_equal(self):
        self.assertEqual(str(self.parse('A1+1')), '(A1) + (1)')

    def test_parse_with_named_range(self):
        # Named ranges get immediately resolved.
        self.assertEqual(str(self.parse('first+1')), '(A1) + (1)')

    def test_parse_function(self):
        self.assertEqual(str(self.parse('SUM(A1:B2)')), 'SUM(A1:B2)')

    def test_parse_function_with_arg_expr(self):
        self.assertEqual(
            str(self.parse('SUM(1, 2, (3+4))')), 'SUM(1, 2, (3) + (4))')

    def test_parse_nested_function(self):
        self.assertEqual(
            str(self.parse('MOD(SUM(A1:B2), 2)')),
            'MOD(SUM(A1:B2), 2)')

    def test_parse_with_parens(self):
        self.assertEqual(str(self.parse('2*(A1+1)')), '(2) * ((A1) + (1))')

    def test_parse_prefix_op(self):
        self.assertEqual(str(self.parse('-(A1+1)')), '- ((A1) + (1))')

    def test_parse_consecutive_unary_negation(self):
        self.assertEqual(str(self.parse('--A1')), '- (- (A1))')

    def test_parse_infix_and_prefix_op(self):
        self.assertEqual(str(self.parse('B1--A1')), '(B1) - (- (A1))')

    def test_parse_percent(self):
        # Should be a proper postfix op, but tokenizer takes care of it right
        # away.
        self.assertEqual(str(self.parse('(A1+1)%')), '((A1) + (1)) * (0.01)')

    def test_parse_open_range_start(self):
        self.assertEqual(str(self.parse(':B2')), ':B2')

    def test_parse_open_range_end(self):
        self.assertEqual(str(self.parse('A1:')), 'A1:')

    @unittest.skip(
        'Not properly supported during shunting. Does not remove func node.')
    def test_parse_with_offser(self):
        self.assertEqual(str(self.parse(':OFFSET(A1, 1, 1)')), ':B2')

    def test_parse_with_parens_mismatched(self):
        with self.assertRaises(SyntaxError):
            self.parse('2*(A1+1')
        with self.assertRaises(IndexError):
            self.parse('2*)')



from .. import testing


class DatedifTest(testing.FunctionalTestCase):
    filename = "DATEDIF.xlsx"

    def test_counta_evaluation_Y(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!C2')
        value = self.evaluator.evaluate('Sheet1!C2')
        self.assertEqual(excel_value, value)

    def test_evaluation_M(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!C3')
        value = self.evaluator.evaluate('Sheet1!C3')
        self.assertEqual(excel_value, value)

    def test_evaluation_D(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!C4')
        value = self.evaluator.evaluate('Sheet1!C4')
        self.assertEqual(excel_value, value)

    def test_evaluation_MD(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!C5')
        value = self.evaluator.evaluate('Sheet1!C5')
        self.assertEqual(excel_value, value)

    def test_evaluation_YM(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!C6')
        value = self.evaluator.evaluate('Sheet1!C6')
        self.assertEqual(excel_value, value)

    def test_evaluation_YD(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!C7')
        value = self.evaluator.evaluate('Sheet1!C7')
        self.assertEqual(excel_value, value)



import copy
import gzip
import jsonpickle
import logging
import os
from dataclasses import dataclass, field

from . import xltypes, reader, parser, tokenizer


@dataclass
class Model():

    cells: dict = field(
        init=False, default_factory=dict, compare=True, hash=True, repr=True)
    formulae: dict = field(
        init=False, default_factory=dict, compare=True, hash=True, repr=True)
    ranges: dict = field(
        init=False, default_factory=dict, compare=True, hash=True, repr=True)
    defined_names: dict = field(
        init=False, default_factory=dict, compare=True, hash=True, repr=True)

    def set_cell_value(self, address, value):
        """Sets a new value for a specified cell."""
        if address in self.defined_names:
            if isinstance(self.defined_names[address], xltypes.XLCell):
                address = self.defined_names[address].address

        if isinstance(address, str):
            if address in self.cells:
                self.cells[address].value = copy.copy(value)
            else:
                self.cells[address] = xltypes.XLCell(address, copy.copy(value))

        elif isinstance(address, xltypes.XLCell):
            if address.address in self.cells:
                self.cells[address.address].value = value
            else:
                self.cells[address.address] = xltypes.XLCell
                (address.address, value)

        else:
            raise TypeError(
                f"Cannot set the cell value for an address of type "
                f"{address}. XLCell or a string is needed."
            )

    def get_cell_value(self, address):
        if address in self.defined_names:
            if isinstance(self.defined_names[address], xltypes.XLCell):
                address = self.defined_names[address].address

        if isinstance(address, str):
            if address in self.cells:
                return self.cells[address].value
            else:
                logging.debug(
                    "Trying to get value for cell {address} but that cell "
                    "doesn't exist.")
                return 0

        elif isinstance(address, xltypes.XLCell):
            if address.address in self.cells:
                return self.cells[address.address].value
            else:
                logging.debug(
                    "Trying to get value for cell {address.address} but "
                    "that cell doesn't exist")
                return 0

        else:
            raise TypeError(
                f"Cannot set the cell value for an address of type "
                f"{address}. XLCell or a string is needed."
            )

    def persist_to_json_file(self, fname):
        """Writes the state to disk.

        Doesn't write the graph directly, but persist all the things that
        provide the ability to re-create the graph.
        """
        output = {
            'cells': self.cells,
            'defined_names': self.defined_names,
            'formulae': self.formulae,
            'ranges': self.ranges,
        }

        file_open = gzip.GzipFile \
            if os.path.splitext(fname)[-1].lower() in ['.gzip', '.gz'] \
            else open

        with file_open(fname, 'wb') as fp:
            fp.write(jsonpickle.encode(output, keys=True).encode())

    def construct_from_json_file(self, fname, build_code=False):
        """Constructs a graph from a state persisted to disk."""

        file_open = gzip.GzipFile \
            if os.path.splitext(fname)[-1].lower() in ['.gzip', '.gz'] \
            else open

        with file_open(fname, "rb") as fp:
            json_bytes = fp.read()

        data = jsonpickle.decode(
            json_bytes, keys=True,
            classes=(
                xltypes.XLCell, xltypes.XLFormula, xltypes.XLRange,
                tokenizer.f_token
            )
        )
        self.cells = data['cells']

        self.defined_names = data['defined_names']
        self.ranges = data['ranges']
        self.formulae = data['formulae']

        if build_code:
            self.build_code()

    def build_code(self):
        """Define the Python code for all cells in the dict of cells."""

        for cell in self.cells:
            if self.cells[cell].formula is not None:
                defined_names = {
                    name: defn.address
                    for name, defn in self.defined_names.items()}
                self.cells[cell].formula.ast = parser.FormulaParser().parse(
                    self.cells[cell].formula.formula, defined_names)

    def __eq__(self, other):

        cells_comparison = []
        for self_cell in self.cells:
            cells_comparison.append(
                self.cells[self_cell] == other.cells[self_cell])

        defined_names_comparison = []
        for self_defined_names in self.defined_names:
            defined_names_comparison.append(
                self.defined_names[self_defined_names]
                    == other.defined_names[self_defined_names])

        return (
            self.__class__ == other.__class__
            and all(cells_comparison)
            and all(defined_names_comparison)
        )


class ModelCompiler:
    """Excel Workbook Data Model Compiler

    Factory class responsible for taking Microsoft Excel cells and named_range
    and create a model represented by a network graph that can be serialized
    to disk, and executed independently of Excel.
    """

    def __init__(self):
        self.model = Model()

    def read_excel_file(self, file_name):
        archive = reader.Reader(file_name)
        archive.read()
        return archive

    def parse_archive(self, archive, ignore_sheets=[], ignore_hidden=False):
        self.model.cells, self.model.formulae, self.model.ranges = \
            archive.read_cells(ignore_sheets, ignore_hidden)
        self.defined_names = archive.read_defined_names(
            ignore_sheets, ignore_hidden)
        self.build_defined_names()
        self.link_cells_to_defined_names()
        self.build_ranges()

    def read_and_parse_archive(
            self, file_name=None, ignore_sheets=[], ignore_hidden=False,
            build_code=True
    ):
        archive = self.read_excel_file(file_name)
        self.parse_archive(archive, ignore_sheets=ignore_sheets)

        if build_code:
            self.model.build_code()

        return self.model

    def read_and_parse_dict(
            self, input_dict, default_sheet="Sheet1", build_code=True):
        for item in input_dict:
            if "!" in item:
                cell_address = item
            else:
                cell_address = "{}!{}".format(default_sheet, item)

            if (
                    not isinstance(input_dict[item], (float, int))
                    and input_dict[item][0] == '='
            ):
                formula = xltypes.XLFormula(
                    input_dict[item],
                    sheet_name=default_sheet
                )
                cell = xltypes.XLCell(
                    cell_address, None,
                    formula=formula)
                self.model.cells[cell_address] = cell
                self.model.formulae[cell_address] = cell.formula

            else:
                self.model.cells[cell_address] = xltypes.XLCell(
                    cell_address, input_dict[item])

        self.build_ranges(default_sheet=default_sheet)

        if build_code:
            self.model.build_code()

        return self.model

    def build_defined_names(self):
        """Add defined ranges to model."""
        for name in self.defined_names:
            cell_address = self.defined_names[name]
            cell_address = cell_address.replace('$', '')

            # a cell has an address like; Sheet1!A1
            if ':' not in cell_address:
                if cell_address not in self.model.cells:
                    logging.warning(
                        f"Defined name {name} refers to empty cell "
                        f"{cell_address}. Is not being loaded.")
                    continue

                else:
                    if self.model.cells[cell_address] is not None:
                        self.model.defined_names[name] = \
                            self.model.cells[cell_address]

            else:
                self.model.defined_names[name] = xltypes.XLRange(
                    cell_address, name=name)
                self.model.ranges[cell_address] = \
                    self.model.defined_names[name]

            if (
                    cell_address in self.model.formulae
                    and name not in self.model.formulae
            ):
                self.model.formulae[name] = \
                    self.model.cells[cell_address].formula

    def link_cells_to_defined_names(self):
        for name in self.model.defined_names:
            defn = self.model.defined_names[name]

            if isinstance(defn, xltypes.XLCell):
                self.model.cells[defn.address].defined_names.append(name)

            elif isinstance(defn, xltypes.XLRange):
                if any(isinstance(el, list) for el in defn.cells):
                    for column in defn.cells:
                        for row_address in column:
                            self.model.cells[row_address].defined_names.append(
                                name)
                else:
                    # programmer error
                    message = "This isn't a dim2 array. {}".format(name)
                    logging.error(message)
                    raise Exception(message)
            else:
                message = (
                    f"Trying to link cells for {name}, but got unkown "
                    f"type {type(defn)}"
                )
                logging.error(message)
                raise ValueError(message)

    def build_ranges(self, default_sheet=None):
        for formula in self.model.formulae:
            associated_cells = set()
            for range in self.model.formulae[formula].terms:
                if ":" in range:
                    if "!" not in range:
                        range = "{}!{}".format(default_sheet, range)
                    self.model.ranges[range] = xltypes.XLRange(range, range)
                    associated_cells.update([
                        cell
                        for row in self.model.ranges[range].cells
                            for cell in row  # noqa: E131
                    ])
                else:
                    associated_cells.add(range)

                if range in self.model.ranges:
                    for row in self.model.ranges[range].cells:
                        for cell_address in row:
                            if cell_address not in self.model.cells.keys():
                                self.model.cells[cell_address] = \
                                    xltypes.XLCell(cell_address, '')

            if formula in self.model.cells:
                self.model.cells[formula].formula.associated_cells = \
                    associated_cells

            if formula in self.model.defined_names:
                self.model.defined_names[formula].formula.associated_cells = \
                    associated_cells

            self.model.formulae[formula].associated_cells = associated_cells

    @staticmethod
    def extract(model, focus):
        extracted_model = Model()

        for address in focus:
            if isinstance(address, str) and address in model.cells:
                extracted_model.cells[address] = copy.deepcopy(
                    model.cells[address])

            elif isinstance(address, str) and address in model.defined_names:

                extracted_model.defined_names[address] = defn = copy.deepcopy(
                    model.defined_names[address])

                if isinstance(defn, xltypes.XLCell):
                    extracted_model.cells[defn.address] = copy.deepcopy(
                        model.cells[defn.address])

                elif isinstance(defn, xltypes.XLRange):
                    for row in defn.cells:
                        for column in row:
                            extracted_model.cells[column] = copy.deepcopy(
                                model.cells[column])

        terms_to_copy = []
        for addr, cell in extracted_model.cells.items():
            if cell.formula is not None:
                for term in cell.formula.terms:
                    if (term in extracted_model.cells
                            and cell.formula != model.cells[addr].formula):
                        cell.formula = copy.deepcopy(model.cells[addr].formula)

                    elif term not in extracted_model.cells:
                        terms_to_copy.append(term)

        for term in terms_to_copy:
            extracted_model.cells[term] = copy.deepcopy(model.cells[term])

        extracted_model.build_code()

        return extracted_model



from typing import Tuple, Union

from . import xl, xlerrors, func_xltypes, xlcriteria


@xl.register()
@xl.validate_args
def AVERAGE(
        *numbers: Tuple[func_xltypes.Number]
) -> func_xltypes.Number:
    """Returns the average (arithmetic mean) of the arguments.

    https://support.office.com/en-us/article/
        average-function-047bac88-d466-426c-a32b-8f33eb960cf6
    """
    numbers = xl.flatten(numbers)

    # If no non numeric cells, return zero (is what excel does)
    if len(numbers) < 1:
        return 0

    return sum(numbers) / len(numbers)


@xl.register()
@xl.validate_args
def COUNT(*values) -> func_xltypes.Number:
    """Counts the number of cells that contain numbers, and counts numbers
    within the list of arguments.

    https://support.office.com/en-us/article/
        count-function-a59cd7fc-b623-4d93-87a4-d23bf411294c
    """
    values = xl.flatten(values)
    if not len(values) or values[0] is None:
        raise xlerrors.ValueExcelError('value1 is required')

    if len(values) > 255:
        raise xlerrors.ValueExcelError(
            f"Can only have up to 255 supplimentary arguments. "
            f"Provided: {len(values)}")

    return len(list(filter(func_xltypes.Number.is_type, values)))


@xl.register()
@xl.validate_args
def COUNTA(*values):
    """Counts the number of cells that are not empty in a range.

    https://support.office.com/en-us/article/
        counta-function-7dc98875-d5c1-46f1-9a82-53f3219e2509
    """
    values = xl.flatten(values)
    if not len(values) or values[0] is None:
        raise xlerrors.NullExcelError('value1 is required')

    if len(values) > 256:
        raise xlerrors.ValueExcelError(
            f"Can only have up to 256 arguments. "
            f"Provided: {len(values)}")

    cells = list(filter(lambda x: not func_xltypes.Blank.is_blank(x), values))
    return len(cells)


@xl.register()
@xl.validate_args
def COUNTIF(
    countRange: func_xltypes.XlArray,
    criteria: func_xltypes.XlAnything,
) -> func_xltypes.XlNumber:
    """Counts the number of cells that match a condition.

    https://support.microsoft.com/en-us/office/
        countif-function-e0de10c6-f885-4e71-abb4-1f464816df34
    """
    check = xlcriteria.parse_criteria(criteria)
    countRange = countRange.flat
    return sum([check(val) for val in countRange])


@xl.register()
@xl.validate_args
def COUNTIFS(
    countRange1: func_xltypes.XlArray,
    criteria1: func_xltypes.XlAnything,
    *rangesAndCriteria: Tuple[Union[
        func_xltypes.XlArray, func_xltypes.XlAnything
    ]]
) -> func_xltypes.XlNumber:
    """Counts the number of cells that match multiple conditions.

    https://support.microsoft.com/en-us/office/
        countifs-function-dda3dc6e-f74e-4aee-88bc-aa8c2a866842
    """
    ranges = [countRange1.flat]
    checks = [xlcriteria.parse_criteria(criteria1)]
    rangeLen = len(ranges[0])
    newRange = []
    idx = 0
    for item in rangesAndCriteria:
        if idx == rangeLen:
            checks.append(xlcriteria.parse_criteria(item))
            ranges.append(newRange)
            newRange = []
            idx = 0
        else:
            newRange.append(item)
            idx += 1
    return sum(
        [
            all([cfn(cvals[i]) for i, cfn in enumerate(checks)])
            for cvals in zip(*ranges)
        ]
    )


@xl.register()
@xl.validate_args
def MAX(*numbers: Tuple[func_xltypes.Number]):
    """Returns the largest value in a set of values.

    https://support.office.com/en-us/article/
        max-function-e0012414-9ac8-4b34-9a47-73e662c08098
    """
    # If no non numeric cells, return zero (is what excel does)
    if len(numbers) < 1:
        return 0

    return max(filter(func_xltypes.Number.is_type, numbers))


@xl.register()
@xl.validate_args
def MIN(*numbers: Tuple[func_xltypes.Number]):
    """Returns the smallest number in a set of values.

    https://support.office.com/en-us/article/
        min-function-61635d12-920f-4ce2-a70f-96f202dcc152
    """
    # If no non numeric cells, return zero (is what excel does)
    if len(numbers) < 1:
        return 0

    return min(filter(func_xltypes.Number.is_type, numbers))



"""Representation of a Microsoft Excel formula
"""
from dataclasses import dataclass, field
from openpyxl.utils.cell import column_index_from_string
from typing import List

from . import ast_nodes, tokenizer, utils
from .tokenizer import ExcelParserTokens


class XLType:
    pass


@dataclass
class XLFormula(XLType):
    """Representing an Excel Formula"""

    formula: str = field(compare=True, hash=True, repr=True)

    sheet_name: str = field(default=None, repr=True)
    reference: str = field(default=None, repr=True)
    evaluate: bool = field(default=True, repr=True)
    tokens: List[tokenizer.f_token] = field(
        init=False, default_factory=list, repr=True)
    terms: List[str] = field(init=False, default_factory=list, repr=True)
    associated_cells: set = field(init=False, default_factory=set, repr=True)
    ast: ast_nodes.ASTNode = field(init=False, default=None)

    def __post_init__(self):
        """Supplimentary initialisation."""
        self.tokens = tokenizer.ExcelParser().getTokens(self.formula).items
        for token in self.tokens:
            if (
                    (token.ttype == ExcelParserTokens.TOK_TYPE_OPERAND)
                    and (token.tsubtype == ExcelParserTokens.TOK_SUBTYPE_RANGE)
                    and (token.tvalue not in self.terms)
            ):
                # Make sure we have a full address.
                term = token.tvalue
                if '!' not in term:
                    term = f'{self.sheet_name}!{term}'
                self.terms.append(term)


@dataclass
class XLCell(XLType):
    """Excel Cell"""

    address: str = field(compare=False, repr=True)

    sheet: str = field(compare=True, hash=True, init=False, repr=False)
    row: str = field(compare=False, hash=False, init=False, repr=False)
    row_index: int = field(compare=True, hash=True, init=False, repr=False)
    column: str = field(compare=False, hash=False, init=False, repr=False)
    column_index: int = field(compare=True, hash=True, init=False, repr=False)
    value: str = field(compare=False, default=None, repr=True)
    formula: XLFormula = field(
        compare=False, default=None, hash=False, repr=True)
    # These are "back-links" to the defined names in Model.
    defined_names: list = field(compare=False, default_factory=list, repr=True)

    def __post_init__(self):
        self.sheet, self.column, self.row = utils.resolve_address(self.address)
        self.column_index = column_index_from_string(self.column)
        self.row_index = int(self.row)

    def __float__(self):
        return float(self.value)

    def __hash__(self):
        # XXX: Also should not be needed.
        return hash((self.sheet, self.row_index, self.column_index))


@dataclass
class XLRange(XLType):
    """Excel Range"""

    address_str: str = field(compare=False, hash=False, repr=True)
    name: str = field(default=None, compare=False, hash=True, repr=True)

    cells: list = field(init=False, compare=True, hash=False, repr=False)
    sheet: str = field(init=False, default="Sheet1", repr=False)
    value: list = field(default=None, repr=True)

    def __post_init__(self):
        if self.name is None:
            self.name = self.address_str
        self.sheet, self.cells = utils.resolve_ranges(self.address_str)

    @property
    def address(self):
        return self.cells



from .. import testing


class InformationTest(testing.FunctionalTestCase):
    filename = "INFORMATION.xlsx"

    def test_isblank_evaluation_A10(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A10')
        value = self.evaluator.evaluate('Sheet1!A10')
        self.assertEqual(excel_value, value)

    def test_isblank_evaluation_B10(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!B10')
        value = self.evaluator.evaluate('Sheet1!B10')
        self.assertEqual(excel_value, value)

    def test_isblank_evaluation_C10(self):
        # C10 has a space.
        excel_value = self.evaluator.get_cell_value('Sheet1!C10')
        value = self.evaluator.evaluate('Sheet1!C10')
        self.assertEqual(excel_value, value)

    def test_istext_evaluation_D10(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!D10')
        value = self.evaluator.evaluate('Sheet1!D10')
        self.assertEqual(excel_value, value)

    def test_istext_evaluation_E10(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!E10')
        value = self.evaluator.evaluate('Sheet1!E10')
        self.assertEqual(excel_value, value)

    def test_istext_evaluation_F10(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!F10')
        value = self.evaluator.evaluate('Sheet1!F10')
        self.assertEqual(excel_value, value)

    def test_na_evaluation_G1(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!G1')
        value = self.evaluator.evaluate('Sheet1!G1')
        self.assertEqual(excel_value, value)

    def test_isna_evaluation_G10(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!G10')
        value = self.evaluator.evaluate('Sheet1!G10')
        self.assertEqual(excel_value, value)

    def test_isna_evaluation_H10(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!H10')
        value = self.evaluator.evaluate('Sheet1!H10')
        self.assertEqual(excel_value, value)



import inspect

from xlcalculator.xlfunctions import (
    xl,
    xlerrors,
    math,
    operator,
    text,
    func_xltypes
)

from . import utils

PREFIX_OP_TO_FUNC = {
    '-': operator.OP_NEG,
}

POSTFIX_OP_TO_FUNC = {
    '%': operator.OP_PERCENT,
}

INFIX_OP_TO_FUNC = {
    "*": operator.OP_MUL,
    "/": operator.OP_DIV,
    "+": operator.OP_ADD,
    "-": operator.OP_SUB,
    "^": math.POWER,
    "&": text.CONCAT,
    "=": operator.OP_EQ,
    "<>": operator.OP_NE,
    ">": operator.OP_GT,
    "<": operator.OP_LT,
    ">=": operator.OP_GE,
    "<=": operator.OP_LE,
}


class EvalContext:

    cells = None
    ranges = None
    namespace = None
    seen = None
    ref = None
    refsheet = None
    sheet = None

    def __init__(self, namespace=None, ref=None, seen=None):
        self.seen = seen if seen is not None else []
        self.namespace = namespace if namespace is not None else xl.FUNCTIONS
        self.ref = ref
        self.sheet = self.refsheet = ref.split("!")[0]

    def eval_cell(self, addr):
        raise NotImplementedError()

    def set_sheet(self, sheet=None):
        if sheet is None:
            self.sheet = self.refsheet
        else:
            self.sheet = sheet


class ASTNode(object):
    """A generic node in the AST"""

    def __init__(self, token):
        self.token = token

    @property
    def tvalue(self):
        return self.token.tvalue

    @property
    def ttype(self):
        return self.token.ttype

    @property
    def tsubtype(self):
        return self.token.tsubtype

    def __eq__(self, other):
        return self.token == other.token

    def eval(self, context):
        raise NotImplementedError(f'`eval()` of {self}')

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} "
            f"tvalue: {repr(self.tvalue)}, "
            f"ttype: {self.ttype}, "
            f"tsubtype: {self.tsubtype}"
            f">"
        )

    def __str__(self):
        return str(self.tvalue)

    def __iter__(self):
        yield self


class OperandNode(ASTNode):

    def eval(self, context):
        if self.tsubtype == "logical":
            return func_xltypes.Boolean.cast(self.tvalue)
        elif self.tsubtype == 'text':
            return func_xltypes.Text(self.tvalue)
        elif self.tsubtype == 'error':
            if self.tvalue in xlerrors.ERRORS_BY_CODE:
                return xlerrors.ERRORS_BY_CODE[self.tvalue](
                    f'Error in cell ${context.ref}')
            return xlerrors.ExcelError(
                self.tvalue, f'Error in cell ${context.ref}')
        else:
            return func_xltypes.Number.cast(self.tvalue)

    def __str__(self):
        if self.tsubtype == "logical":
            return self.tvalue.title()
        elif self.tsubtype == "text":
            return '"' + self.tvalue.replace('"', '\\"') + '"'
        return str(self.tvalue)


class RangeNode(OperandNode):
    """Represents a spreadsheet cell, range, named_range."""

    def get_cells(self):
        cells = utils.resolve_ranges(self.tvalue, default_sheet='')[1]
        return cells[0] if len(cells) == 1 else cells

    @property
    def address(self):
        return self.tvalue

    def full_address(self, context):
        addr = self.address
        if '!' not in addr:
            addr = f'{context.sheet}!{addr}'
        return addr

    def eval(self, context):
        addr = self.full_address(context)

        if addr in context.ranges:
            range_cells = [
                [
                    context.eval_cell(addr)
                    for addr in range_row
                ]
                for range_row in context.ranges[addr].cells
            ]
            context.ranges[addr].value = data = func_xltypes.Array(range_cells)
            return data

        value = context.eval_cell(addr)
        context.set_sheet()
        return value


class OperatorNode(ASTNode):

    def __init__(self, token):
        super().__init__(token)
        self.left = None
        self.right = None

    def eval(self, context):
        if self.ttype == 'operator-prefix':
            assert self.left is None, 'Left operand for prefix operator'
            op = PREFIX_OP_TO_FUNC[self.tvalue]
            return op(self.right.eval(context))

        elif self.ttype == 'operator-infix':
            op = INFIX_OP_TO_FUNC[self.tvalue]
            return op(
                self.left.eval(context),
                self.right.eval(context),
            )
        elif self.ttype == 'operator-postfix':
            assert self.right is None, 'Right operand for postfix operator'
            op = POSTFIX_OP_TO_FUNC[self.tvalue]
            return op(self.left.eval(context))
        else:
            raise ValueError(f'Invalid operator type: {self.ttype}')

    def __str__(self):
        left = f'({self.left}) ' if self.left is not None else ''
        right = f' ({self.right})' if self.right is not None else ''
        return f'{left}{self.tvalue}{right}'

    def __iter__(self):
        # Return node in resolution order.
        yield self.left
        yield self.right
        yield self


class FunctionNode(ASTNode):
    """AST node representing a function call"""

    def __init__(self, token):
        super().__init__(token)
        self.args = None

    def eval(self, context):
        func_name = self.tvalue.upper()
        # 1. Remove the BBB namespace, since we are just supporting
        #    everything in one large one.
        func_name = func_name.replace('_XLFN.', '')
        # 2. Look up the function to use.
        func = context.namespace[func_name]
        # 3. Prepare arguments.
        sig = inspect.signature(func)
        bound = sig.bind(*self.args)
        args = []
        for pname, pvalue in list(bound.arguments.items()):
            param = sig.parameters[pname]
            ptype = param.annotation
            if ptype == func_xltypes.XlExpr:
                args.append(func_xltypes.Expr(
                    pvalue.eval, (context,), ref=context.ref, ast=pvalue
                ))
            elif (param.kind == param.VAR_POSITIONAL
                  and func_xltypes.XlExpr in getattr(ptype, '__args__', [])):
                args.extend([
                    func_xltypes.Expr(
                        pitem.eval, (context,), ref=context.ref, ast=pitem
                    )
                    for pitem in pvalue
                ])
            elif (param.kind == param.VAR_POSITIONAL):
                args.extend([
                    pitem.eval(context) for pitem in pvalue
                ])
            else:
                args.append(pvalue.eval(context))
        # 4. Run function and return result.
        return func(*args)

    def __str__(self):
        args = ', '.join(str(arg) for arg in self.args)
        return f'{self.tvalue}({args})'

    def __iter__(self):
        # Return node in resolution order.
        for arg in self.args:
            yield arg
        yield self



from .. import testing


class ChooseTest(testing.FunctionalTestCase):
    filename = "choose.xlsx"

    def test_evaluation_A1(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A1')
        value_00 = self.evaluator.evaluate('Sheet1!A1')
        self.assertEqual(excel_value, value_00)

    def test_evaluation_A2(self):
        excel_value = self.evaluator.get_cell_value('Sheet1!A2')
        value_01 = self.evaluator.evaluate('Sheet1!A2')
        self.assertEqual(excel_value, value_01)

    def test_evaluation_A3(self):
        excel_value = [[1, 2, 3]]
        value_00 = self.evaluator.evaluate('Sheet1!A3')
        self.assertEqual(excel_value, value_00.values.tolist())



import unittest


class VDBTest(unittest.TestCase):
    pass

    # def test_evaluation_F23(self):
    #     excel_value = self.evaluator.get_cell_value('Sheet1!F23')
    #     value = self.evaluator.evaluate('Sheet1!F23')
    #     self.assertEqual( excel_value, value )
    #

    # def test_evaluation_F24(self):
    #     excel_value = self.evaluator.get_cell_value('Sheet1!F24')
    #     value = self.evaluator.evaluate('Sheet1!F24')
    #     self.assertEqual( excel_value, value )
    #
    #
    # def test_evaluation_F27(self):
    #     excel_value = self.evaluator.get_cell_value('Sheet1!F27')
    #     value = self.evaluator.evaluate('Sheet1!F27')
    #     self.assertEqual( excel_value, value )
    #
    #
    # def test_evaluation_F28(self):
    #     excel_value = self.evaluator.get_cell_value('Sheet1!F28')
    #     value = self.evaluator.evaluate('Sheet1!F28')
    #     self.assertEqual( excel_value, value )



import os
import setuptools


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


TESTS_REQUIRE = [
    'coverage',
    'flake8',
    'mock',
    'pytest',
    'pytest-cov',
]


setuptools.setup(
    name="xlcalculator",
    version='0.3.0',
    author="Bradley van Ree",
    author_email="brads@bradbase.net",
    description="Converts MS Excel formulas to Python and evaluates them.",
    long_description=(
        read('README.rst')
        + '\n\n' +
        read('CHANGES.rst')
        ),
    url="https://github.com/bradbase/xlcalculator",
    packages=setuptools.find_packages(),
    license="MIT",
    keywords=['xls',
        'Excel',
        'spreadsheet',
        'workbook',
        'data analysis',
        'analysis'
        'reading excel',
        'excel formula',
        'excel formulas',
        'excel equations',
        'excel equation',
        'formula',
        'formulas',
        'equation',
        'equations',
        'timeseries',
        'time series',
        'research',
        'scenario analysis',
        'scenario',
        'modelling',
        'model',
        'unit testing',
        'testing',
        'audit',
        'calculation',
        'evaluation',
        'data science',
        'openpyxl'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Financial and Insurance Industry",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Office/Business :: Financial :: Spreadsheet",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Testing :: Unit",
        "Topic :: Utilities",
    ],
    install_requires=[
        'jsonpickle',
        'numpy',
        'pandas',
        'openpyxl',
        'numpy-financial',
        'yearfrac==0.4.4',
        'mock',
        'scipy'
    ],
    extras_require=dict(
        test=TESTS_REQUIRE,
        build=[
            'pip-tools',
        ],
    ),
    python_requires='>=3.7',
    tests_require=TESTS_REQUIRE,
    include_package_data=True,
    zip_safe=False,
)



import datetime
import mock
import unittest

from xlcalculator.xlfunctions import date, xlerrors, func_xltypes, utils

dt = datetime.datetime


class DateModuleTest(unittest.TestCase):

    def test_DATE(self):
        self.assertEqual(date.DATE(2000, 1, 1), dt(2000, 1, 1))
        self.assertEqual(date.DATE(2008, 11, 3), dt(2008, 11, 3))
        self.assertEqual(date.DATE(2024, 1, 1), dt(2024, 1, 1))
        self.assertEqual(date.DATE(2025, 1, 1), dt(2025, 1, 1))
        self.assertEqual(date.DATE(2026, 1, 1), dt(2026, 1, 1))
        self.assertEqual(date.DATE(1911, 1, 15), dt(1911, 1, 15))

    def test_DATE_with_short_year(self):
        self.assertEqual(date.DATE(99, 1, 1), dt(1999, 1, 1))

    def test_DATE_with_string_components(self):
        self.assertEqual(date.DATE('2026', '1', '1'), dt(2026, 1, 1))

    def test_DATE_year_must_be_number(self):
        self.assertIsInstance(date.DATE('bad', 1, 1), xlerrors.ValueExcelError)

    def test_DATE_month_must_be_number(self):
        self.assertIsInstance(
            date.DATE(2000, 'bad', 1), xlerrors.ValueExcelError)

    def test_DATE_day_must_be_integer(self):
        self.assertIsInstance(
            date.DATE(2000, 1, 'bad'), xlerrors.ValueExcelError)

    def test_DATE_year_must_be_positive(self):
        self.assertIsInstance(date.DATE(-1, 1, 1), xlerrors.NumExcelError)

    def test_DATE_year_must_have_less_than_10000(self):
        self.assertIsInstance(date.DATE(10000, 1, 1), xlerrors.NumExcelError)

    def test_DATE_result_must_be_positive(self):
        self.assertIsInstance(date.DATE(1900, 1, -1), xlerrors.NumExcelError)

    def test_DATE_not_stricly_positive_month_substracts(self):
        self.assertEqual(date.DATE(2009, -1, 1), dt(2008, 11, 1))

    def test_DATE_not_stricly_positive_day_substracts(self):
        self.assertEqual(date.DATE(2009, 1, -1), dt(2008, 12, 30))

    def test_DATE_month_superior_to_12_change_year(self):
        self.assertEqual(date.DATE(2009, 14, 1), dt(2010, 2, 1))

    def test_DATE_day_superior_to_365_change_year(self):
        self.assertEqual(date.DATE(2009, 1, 400), dt(2010, 2, 4))

    def test_DATEDIF_year(self):
        start_date = date.DATE(2011, 1, 1)
        end_date = date.DATE(2011, 12, 31)

        self.assertEqual(date.DATEDIF(start_date, end_date, 'Y'), 0)

        start_date = date.DATE(2011, 1, 1)
        end_date = date.DATE(2012, 12, 31)
        self.assertEqual(date.DATEDIF(start_date, end_date, 'Y'), 1)

    def test_DATEDIF_month(self):
        start_date = date.DATE(2011, 1, 1)
        end_date = date.DATE(2011, 12, 31)

        self.assertEqual(date.DATEDIF(start_date, end_date, 'M'), 11)

    def test_DATEDIF_day(self):
        start_date = date.DATE(2001, 6, 1)
        end_date = date.DATE(2002, 8, 15)

        self.assertEqual(date.DATEDIF(start_date, end_date, 'D'), 440)

    def test_DATEDIF_MD(self):
        start_date = date.DATE(2011, 1, 1)
        end_date = date.DATE(2011, 12, 31)

        self.assertEqual(date.DATEDIF(start_date, end_date, 'MD'), 30)

    def test_DATEDIF_YM(self):
        start_date = date.DATE(2011, 1, 1)
        end_date = date.DATE(2011, 12, 31)

        self.assertEqual(date.DATEDIF(start_date, end_date, 'YM'), 11)

    def test_DATEDIF_YD(self):
        start_date = date.DATE(2001, 6, 1)
        end_date = date.DATE(2002, 8, 15)

        self.assertEqual(date.DATEDIF(start_date, end_date, 'YD'), 75)

    def test_DATEDIF_M_start_beyond_end(self):
        start_date = date.DATE(2002, 8, 15)
        end_date = date.DATE(2001, 6, 1)

        self.assertIsInstance(date.DATEDIF(start_date, end_date, 'M'),
                              xlerrors.NumExcelError)

    def test_DAY(self):
        sample_date = date.DATE(1911, 4, 15)
        serial_number = int(sample_date)

        self.assertEqual(date.DAY(serial_number), 15)

    def test_DAYS(self):
        end_date = date.DATE(2011, 12, 31)
        start_date = date.DATE(2011, 1, 1)

        self.assertEqual(date.DAYS(end_date, start_date), 364)

    def test_EDATE(self):
        the_date = date.DATE(2011, 1, 15)
        target_serial_number = utils.datetime_to_number(dt(2011, 2, 15))
        self.assertEqual(date.EDATE(the_date, 1), target_serial_number)

    def test_EDATE_must_be_greater_than_epoch(self):
        the_date = date.DATE(1901, 1, 15)
        self.assertIsInstance(date.EDATE(the_date, -24),
                              xlerrors.NumExcelError)

    def test_EOMONTH(self):
        the_date = date.DATE(2011, 1, 1)
        target_serial_number = utils.datetime_to_number(dt(2011, 2, 28))
        self.assertEqual(date.EOMONTH(the_date, 1), target_serial_number)

    def test_EOMONTH_must_be_greater_than_epoch(self):
        the_date = date.DATE(1901, 1, 1)
        self.assertIsInstance(date.EOMONTH(the_date, -24),
                              xlerrors.NumExcelError)

    def test_ISOWEEKNUM(self):
        sample_date = date.DATE(2012, 3, 9)

        self.assertEqual(date.ISOWEEKNUM(sample_date), 10)

    def test_MONTH(self):
        sample_date = date.DATE(1911, 4, 15)
        serial_number = int(sample_date)

        self.assertEqual(date.MONTH(serial_number), 4)

    def test_NOW(self):
        with mock.patch.object(date, 'now', lambda: dt(2000, 1, 1, 12, 30)):
            self.assertEqual(
                date.NOW(), utils.datetime_to_number(dt(2000, 1, 1, 12, 30))
            )

    def test_TODAY(self):
        with mock.patch.object(date, 'now', lambda: dt(2000, 1, 1, 12, 30)):
            self.assertEqual(
                date.TODAY(), utils.datetime_to_number(dt(2000, 1, 1))
            )

    def test_WEEKDAY_return_type_omitted(self):
        sample_date = date.DATE(2008, 2, 14)
        serial_number = int(sample_date)

        self.assertEqual(date.WEEKDAY(serial_number), 5)

    def test_WEEKDAY_return_type_1(self):
        sample_date = date.DATE(2008, 2, 14)
        serial_number = int(sample_date)

        self.assertEqual(date.WEEKDAY(serial_number, 1), 5)

    def test_WEEKDAY_return_type_2(self):
        sample_date = date.DATE(2008, 2, 14)
        serial_number = int(sample_date)

        self.assertEqual(date.WEEKDAY(serial_number, 2), 4)

    def test_WEEKDAY_return_type_3(self):
        sample_date = date.DATE(2008, 2, 14)
        serial_number = int(sample_date)

        self.assertEqual(date.WEEKDAY(serial_number, 3), 3)

    def test_WEEKDAY_return_type_11(self):
        sample_date = date.DATE(2008, 2, 14)
        serial_number = int(sample_date)

        self.assertEqual(date.WEEKDAY(serial_number, 11), 4)

    def test_WEEKDAY_return_type_12(self):
        sample_date = date.DATE(2008, 2, 14)
        serial_number = int(sample_date)

        self.assertEqual(date.WEEKDAY(serial_number, 12), 3)

    def test_WEEKDAY_return_type_13(self):
        sample_date = date.DATE(2008, 2, 14)
        serial_number = int(sample_date)

        self.assertEqual(date.WEEKDAY(serial_number, 13), 2)

    def test_WEEKDAY_return_type_14(self):
        sample_date = date.DATE(2008, 2, 14)
        serial_number = int(sample_date)

        self.assertEqual(date.WEEKDAY(serial_number, 14), 1)

    def test_WEEKDAY_return_type_15(self):
        sample_date = date.DATE(2008, 2, 14)
        serial_number = int(sample_date)

        self.assertEqual(date.WEEKDAY(serial_number, 15), 7)

    def test_WEEKDAY_return_type_16(self):
        sample_date = date.DATE(2008, 2, 14)
        serial_number = int(sample_date)

        self.assertEqual(date.WEEKDAY(serial_number, 16), 6)

    def test_WEEKDAY_return_type_17(self):
        sample_date = date.DATE(2008, 2, 14)
        serial_number = int(sample_date)

        self.assertEqual(date.WEEKDAY(serial_number, 17), 5)

    def test_WEEKDAY_error(self):
        sample_date = date.DATE(2008, 2, 14)
        serial_number = int(sample_date)

        self.assertIsInstance(date.WEEKDAY(serial_number, 0),
                              xlerrors.NumExcelError)
        self.assertIsInstance(date.WEEKDAY(serial_number, 4),
                              xlerrors.NumExcelError)
        self.assertIsInstance(date.WEEKDAY(serial_number, 5),
                              xlerrors.NumExcelError)
        self.assertIsInstance(date.WEEKDAY(serial_number, 6),
                              xlerrors.NumExcelError)
        self.assertIsInstance(date.WEEKDAY(serial_number, 7),
                              xlerrors.NumExcelError)
        self.assertIsInstance(date.WEEKDAY(serial_number, 8),
                              xlerrors.NumExcelError)
        self.assertIsInstance(date.WEEKDAY(serial_number, 9),
                              xlerrors.NumExcelError)
        self.assertIsInstance(date.WEEKDAY(serial_number, 10),
                              xlerrors.NumExcelError)
        self.assertIsInstance(date.WEEKDAY(serial_number, 18),
                              xlerrors.NumExcelError)

    def test_YEAR(self):
        sample_date_2008 = date.DATE(2008, 5, 7)
        serial_number_2008 = int(sample_date_2008)
        self.assertEqual(date.YEAR(serial_number_2008), 2008)

        sample_date_2010 = date.DATE(2010, 5, 7)
        serial_number_2010 = int(sample_date_2010)
        self.assertEqual(date.YEAR(serial_number_2010), 2010)

    def test_YEAR_before_epoch(self):
        # is text as a "date" entry into Excel of a date pre-epoch
        # is stored as datatype General
        sample_date = func_xltypes.XlText('12/31/1899')

        self.assertIsInstance(
            date.YEAR(sample_date), xlerrors.ValueExcelError)

    def test_YEAR_after_9999(self):
        # is text as a "date" entry into Excel of a date pre-epoch
        # is stored as datatype General
        sample_date = func_xltypes.XlText('1/1/10000')

        self.assertIsInstance(
            date.YEAR(sample_date), xlerrors.ValueExcelError)

    def test_YEARFRAC_start_date_must_be_datetime(self):
        self.assertIsInstance(
            date.YEARFRAC('bad', 1), xlerrors.ValueExcelError)

    def test_YEARFRAC_end_date_must_be_datetime(self):
        self.assertIsInstance(
            date.YEARFRAC(1, 'bad'), xlerrors.ValueExcelError)

    def test_YEARFRAC_start_date_must_after_epoch(self):
        self.assertIsInstance(
            date.YEARFRAC(dt(1800, 1, 1), 0), xlerrors.ValueExcelError)

    def test_YEARFRAC_end_date_must_be_after_epoch(self):
        self.assertIsInstance(
            date.YEARFRAC(dt(1900, 1, 1), -1), xlerrors.ValueExcelError)

    def test_YEARFRAC_basis_must_be_between_0_and_4(self):
        self.assertIsInstance(
            date.YEARFRAC(1, 2, 5), xlerrors.ValueExcelError)
        self.assertIsInstance(
            date.YEARFRAC(1, 2, -1), xlerrors.ValueExcelError)

    def test_YEARFRAC_yearfrac_basis_0(self):
        # From Excel docs.
        self.assertAlmostEqual(
            date.YEARFRAC(dt(2012, 1, 1), dt(2012, 7, 30), 0),
            0.58055556)
        self.assertAlmostEqual(
            date.YEARFRAC(dt(2008, 1, 1), dt(2015, 4, 20)),
            7.30277777777778)
        self.assertAlmostEqual(
            date.YEARFRAC(dt(2008, 1, 1), dt(2015, 4, 20), 0),
            7.30277777777778)
        self.assertAlmostEqual(
            date.YEARFRAC(dt(2024, 1, 1), dt(2025, 1, 1), 0), 1)

    def test_YEARFRAC_yearfrac_basis_1(self):
        # From Excel docs.
        self.assertAlmostEqual(
            date.YEARFRAC(dt(2012, 1, 1), dt(2012, 7, 30), 1),
            0.57650273)

    def test_YEARFRAC_yearfrac_basis_2(self):
        self.assertAlmostEqual(
            date.YEARFRAC(dt(2008, 1, 1), dt(2015, 4, 20), 2),
            7.405555556)
        self.assertAlmostEqual(
            date.YEARFRAC(dt(2024, 1, 1), dt(2025, 1, 1), 2),
            1.01666666666667)

    def test_YEARFRAC_yearfrac_basis_3(self):
        # From Excel docs.
        self.assertAlmostEqual(
            date.YEARFRAC(dt(2012, 1, 1), dt(2012, 7, 30), 3),
            0.57808219)
        self.assertAlmostEqual(
            date.YEARFRAC(dt(2008, 1, 1), dt(2015, 4, 20), 3),
            7.304109589)
        self.assertAlmostEqual(
            date.YEARFRAC(dt(2024, 1, 1), dt(2025, 1, 1), 3),
            1.0027397260274)

    def test_YEARFRAC_yearfrac_basis_4(self):
        self.assertAlmostEqual(
            date.YEARFRAC(dt(2008, 1, 1), dt(2015, 4, 20), 4),
            7.302777778)
        self.assertAlmostEqual(
            date.YEARFRAC(dt(2024, 1, 1), dt(2025, 1, 1), 4), 1)

    def test_YEARFRAC_yearfrac_inverted(self):
        self.assertAlmostEqual(
            date.YEARFRAC(dt(2015, 4, 20), dt(2008, 1, 1)),
            date.YEARFRAC(dt(2008, 1, 1), dt(2015, 4, 20)))



import collections
import re
from openpyxl.utils.cell import COORD_RE, SHEET_TITLE
from openpyxl.utils.cell import range_boundaries, get_column_letter

MAX_COL = 18278
MAX_ROW = 1048576


def resolve_sheet(sheet_str):
    sheet_str = sheet_str.strip()
    sheet_match = re.match(SHEET_TITLE.strip(), sheet_str + '!')
    if sheet_match is None:
        # Internally, sheets are not properly quoted, so consider the entire
        # string.
        return sheet_str

    return sheet_match.group("quoted") or sheet_match.group("notquoted")


def resolve_address(addr):
    # Addresses without sheet name are not supported.
    sheet_str, addr_str = addr.split('!')
    sheet = resolve_sheet(sheet_str)
    coord_match = COORD_RE.split(addr_str)
    col, row = coord_match[1:3]
    return sheet, col, row


def resolve_ranges(ranges, default_sheet='Sheet1'):
    sheet = None
    range_cells = collections.defaultdict(set)
    for rng in ranges.split(','):
        # Handle sheets in range.
        if '!' in rng:
            sheet_str, rng = rng.split('!')
            rng_sheet = resolve_sheet(sheet_str)
            if sheet is not None and sheet != rng_sheet:
                raise ValueError(
                    f'Got multiple different sheets in ranges: '
                    f'{sheet}, {rng_sheet}'
                )
            sheet = rng_sheet
        min_col, min_row, max_col, max_row = range_boundaries(rng)

        if (
                min_col is None or min_row is None
                or max_col is None or max_row is None
        ):
            continue

        # Excel ranges are boundaries inclusive!
        for row_idx in range(min_row, max_row + 1):
            row_cells = range_cells[row_idx]
            for col_idx in range(min_col, max_col + 1):
                row_cells.add(col_idx)

    # Now convert the internal structure to a matrix of cell addresses.
    sheet = default_sheet if sheet is None else sheet
    sheet_str = sheet + '!' if sheet else ''
    return sheet, [
        [
            f'{sheet_str}{get_column_letter(col_idx)}{row_idx}'
            for col_idx in sorted(row_cells)
        ]
        for row_idx, row_cells in sorted(range_cells.items())
    ]



import unittest

from xlcalculator.xlfunctions import xlerrors


class ExcelErrorTest(unittest.TestCase):

    def test_init(self):
        err = xlerrors.ExcelError('#N/A', 'Not applicable')
        self.assertEqual(err.value, '#N/A')
        self.assertEqual(err.info, 'Not applicable')

    def test_str(self):
        err = xlerrors.ExcelError('#N/A', 'Not applicable')
        self.assertEqual(str(err), '#N/A')

    def test_eq(self):
        err = xlerrors.ExcelError('#N/A', 'Not applicable')
        self.assertEqual(err, err)

    def test_eq_to_string(self):
        err = xlerrors.ExcelError('#N/A', 'Not applicable')
        self.assertEqual(err, '#N/A')


class SpecificExcelErrorTest(unittest.TestCase):

    def test_init(self):
        err = xlerrors.SpecificExcelError('Not applicable')
        self.assertEqual(err.value, None)
        self.assertEqual(err.info, 'Not applicable')


class NullExcelErrorTest(unittest.TestCase):

    def test_error(self):
        err = xlerrors.NullExcelError('Error')
        self.assertEqual(str(err), xlerrors.ERROR_CODE_NULL)


class DivZeroExcelErrorTest(unittest.TestCase):

    def test_error(self):
        err = xlerrors.DivZeroExcelError('Error')
        self.assertEqual(str(err), xlerrors.ERROR_CODE_DIV_ZERO)


class ValueExcelErrorTest(unittest.TestCase):

    def test_error(self):
        err = xlerrors.ValueExcelError('Error')
        self.assertEqual(str(err), xlerrors.ERROR_CODE_VALUE)


class RefExcelErrorTest(unittest.TestCase):

    def test_error(self):
        err = xlerrors.RefExcelError('Error')
        self.assertEqual(str(err), xlerrors.ERROR_CODE_REF)


class NameExcelErrorTest(unittest.TestCase):

    def test_error(self):
        err = xlerrors.NameExcelError('Error')
        self.assertEqual(str(err), xlerrors.ERROR_CODE_NAME)


class NumExcelErrorTest(unittest.TestCase):

    def test_error(self):
        err = xlerrors.NumExcelError('Error')
        self.assertEqual(str(err), xlerrors.ERROR_CODE_NUM)


class NaExcelErrorTest(unittest.TestCase):

    def test_error(self):
        err = xlerrors.NaExcelError('Error')
        self.assertEqual(str(err), xlerrors.ERROR_CODE_NA)



import datetime
import yearfrac
from dateutil.relativedelta import relativedelta
from dateutil import rrule

from . import utils, xl, xlerrors, func_xltypes

# Testing hook.
now = datetime.datetime.now


@xl.register()
@xl.validate_args
def DATE(
        year: func_xltypes.XlNumber,
        month: func_xltypes.XlNumber,
        day: func_xltypes.XlNumber
) -> func_xltypes.XlDateTime:
    """Returns the sequential serial number that represents a particular date.

    https://support.office.com/en-us/article/
        date-function-e36c0c8c-4104-49da-ab83-82328b832349
    """
    if not (0 < year < 9999):
        raise xlerrors.NumExcelError(
            f'Year must be between 1 and 9999, got {year}')

    if year < 1900:
        year = 1900 + year

    # Excel starts counting at 1 and today is inclusive, thus -2
    delta = relativedelta(
        years=year - 1900, months=int(month) - 1, days=int(day) - 1)
    result = utils.EXCEL_EPOCH + delta

    if result <= utils.EXCEL_EPOCH:
        raise xlerrors.NumExcelError(
            f"Date result before {utils.EXCEL_EPOCH}")

    return result


@xl.register()
@xl.validate_args
def DATEDIF(
        start_date: func_xltypes.XlDateTime,
        end_date: func_xltypes.XlDateTime,
        unit: func_xltypes.XlText
) -> func_xltypes.XlNumber:
    """Calculates the number of days, months, or years between two dates.

    https://support.office.com/en-us/article/
        datedif-function-25dba1a4-2812-480b-84dd-8b32a451b35c
    """

    if start_date > end_date:
        raise xlerrors.NumExcelError(
            f'Start date must be before the end date. Got Start: \
                {start_date}, End: {end_date}')

    datetime_start_date = utils.number_to_datetime(int(start_date))
    datetime_end_date = utils.number_to_datetime(int(end_date))

    if str(unit).upper() == 'Y':
        date_list = list(rrule.rrule(rrule.YEARLY,
                                     dtstart=datetime_start_date,
                                     until=datetime_end_date))
        return len(date_list) - 1  # end of day to end of day / "full days"

    elif str(unit).upper() == 'M':
        date_list = list(rrule.rrule(rrule.MONTHLY,
                                     dtstart=datetime_start_date,
                                     until=datetime_end_date))
        return len(date_list) - 1  # end of day to end of day / "full days"

    elif str(unit).upper() == 'D':
        date_list = list(rrule.rrule(rrule.DAILY,
                                     dtstart=datetime_start_date,
                                     until=datetime_end_date))
        return len(date_list) - 1  # end of day to end of day / "full days"

    elif str(unit).upper() == 'MD':
        modified_datetime_start_date = datetime_start_date.replace(year=1900,
                                                                   month=1)
        modified_datetime_end_date = datetime_end_date.replace(year=1900,
                                                               month=1)
        date_list = list(rrule.rrule(rrule.DAILY,
                                     dtstart=modified_datetime_start_date,
                                     until=modified_datetime_end_date))
        return len(date_list) - 1  # end of day to end of day / "full days"

    elif str(unit).upper() == 'YM':
        modified_datetime_start_date = datetime_start_date.replace(year=1900,
                                                                   day=1)
        modified_datetime_end_date = datetime_end_date.replace(year=1900,
                                                               day=1)
        date_list = list(rrule.rrule(rrule.MONTHLY,
                                     dtstart=modified_datetime_start_date,
                                     until=modified_datetime_end_date))
        return len(date_list) - 1  # end of day to end of day / "full days"

    elif str(unit).upper() == 'YD':
        modified_datetime_start_date = datetime_start_date.replace(year=1900)
        modified_datetime_end_date = datetime_end_date.replace(year=1900)
        date_list = list(rrule.rrule(rrule.DAILY,
                                     dtstart=modified_datetime_start_date,
                                     until=modified_datetime_end_date))
        return len(date_list) - 1  # end of day to end of day / "full days"


@xl.register()
@xl.validate_args
def DAY(
        serial_number: func_xltypes.XlNumber
) -> func_xltypes.XlNumber:
    """Returns the day of a date, represented by a serial number. The day is
    given as an integer ranging from 1 to 31.

    https://support.microsoft.com/en-us/office/
        day-function-8a7d1cbb-6c7d-4ba1-8aea-25c134d03101
    """

    date = utils.number_to_datetime(int(serial_number))
    return int(date.strftime("%d"))


@xl.register()
@xl.validate_args
def DAYS(
        end_date: func_xltypes.XlDateTime,
        start_date: func_xltypes.XlDateTime
) -> func_xltypes.XlNumber:
    """Returns the number of days between two dates.

    https://support.microsoft.com/en-us/office/
        days-function-57740535-d549-4395-8728-0f07bff0b9df
    """

    days = end_date - start_date
    return days


@xl.register()
@xl.validate_args
def EDATE(
        start_date: func_xltypes.XlDateTime,
        months: func_xltypes.XlNumber
) -> func_xltypes.XlDateTime:
    """Returns the serial number that represents the date that is the
    indicated number of months before or after a specified date
    (the start_date)

    https://support.office.com/en-us/article/
        edate-function-3c920eb2-6e66-44e7-a1f5-753ae47ee4f5
    """
    delta = relativedelta(months=int(months))
    edate = utils.number_to_datetime(int(start_date)) + delta

    if edate <= utils.EXCEL_EPOCH:
        raise xlerrors.NumExcelError(
            f"Date result before {utils.EXCEL_EPOCH}")

    return utils.datetime_to_number(edate)


@xl.register()
@xl.validate_args
def EOMONTH(
        start_date: func_xltypes.XlDateTime,
        months: func_xltypes.XlNumber
) -> func_xltypes.XlNumber:
    """Returns the serial number for the last day of the month that is the
    indicated number of months before or after start_date.

    https://support.office.com/en-us/article/
        eomonth-function-7314ffa1-2bc9-4005-9d66-f49db127d628
    """
    delta = relativedelta(months=int(months))
    edate = utils.number_to_datetime(int(start_date)) + delta

    if edate <= utils.EXCEL_EPOCH:
        raise xlerrors.NumExcelError(
            f"Date result before {utils.EXCEL_EPOCH}")

    eomonth = edate + relativedelta(day=31)

    return utils.datetime_to_number(eomonth)


@xl.register()
@xl.validate_args
def ISOWEEKNUM(
        date: func_xltypes.XlDateTime
) -> func_xltypes.XlNumber:
    """Returns number of the ISO week number of the year for a given date.

    https://support.microsoft.com/en-us/office/
        isoweeknum-function-1c2d0afe-d25b-4ab1-8894-8d0520e90e0e
    """

    datetime_date = utils.number_to_datetime(int(date))
    isoweeknum = datetime_date.isocalendar()[1]
    return isoweeknum


@xl.register()
@xl.validate_args
def MONTH(
        serial_number: func_xltypes.XlNumber
) -> func_xltypes.XlNumber:
    """Returns the month of a date represented by a serial number. The month
    is given as an integer, ranging from 1 (January) to 12 (December).

    https://support.microsoft.com/en-us/office/
        month-function-579a2881-199b-48b2-ab90-ddba0eba86e8
    """

    date = utils.number_to_datetime(int(serial_number))
    return int(date.strftime("%m"))


@xl.register()
def NOW() -> func_xltypes.XlDateTime:
    """Returns the serial number of the current date and time.

    https://support.office.com/en-us/article/
        today-function-5eb3078d-a82c-4736-8930-2f51a028fdd9
    """
    return utils.datetime_to_number(now())


@xl.register()
def TODAY() -> func_xltypes.XlDateTime:
    """Returns the serial number of the current date.

    https://support.office.com/en-us/article/
        today-function-5eb3078d-a82c-4736-8930-2f51a028fdd9
    """
    date_and_time = now()
    date = date_and_time.replace(
        hour=0, minute=0, second=0, microsecond=0)
    return utils.datetime_to_number(date)


@xl.register()
@xl.validate_args
def WEEKDAY(
        serial_number: func_xltypes.XlNumber,
        return_type: func_xltypes.XlNumber = None
) -> func_xltypes.XlNumber:
    """Returns the day of the week corresponding to a date. The day is given
    as an integer, ranging from 1 (Sunday) to 7 (Saturday), by default.

    https://support.microsoft.com/en-us/office/
        weekday-function-60e44483-2ed1-439f-8bd0-e404c190949a
    """

    date = utils.number_to_datetime(int(serial_number))

    if return_type is None:
        # Numbers 1 (Sunday) through 7 (Saturday)
        weekDays = (2, 3, 4, 5, 6, 7, 1)
        return weekDays[date.weekday()]

    elif int(return_type) == 1:
        # Numbers 1 (Sunday) through 7 (Saturday)
        weekDays = (2, 3, 4, 5, 6, 7, 1)
        return weekDays[date.weekday()]

    # weekday() is 0 based, starting on a Monday
    elif int(return_type) == 2:
        # Numbers 1 (Monday) through 7 (Sunday)
        weekDays = (1, 2, 3, 4, 5, 6, 7)
        return weekDays[date.weekday()]

    elif int(return_type) == 3:
        # Numbers 0 (Monday) through 6 (Sunday)
        weekDays = (0, 1, 2, 3, 4, 5, 6)
        return weekDays[date.weekday()]

    elif int(return_type) == 11:
        # Numbers 1 (Monday) through 7 (Sunday)
        weekDays = (1, 2, 3, 4, 5, 6, 7)
        return weekDays[date.weekday()]

    elif int(return_type) == 12:
        # Numbers 1 (Tuesday) through 7 (Monday)
        weekDays = (7, 1, 2, 3, 4, 5, 6)
        return weekDays[date.weekday()]

    elif int(return_type) == 13:
        # Numbers 1 (Wednesday) through 7 (Tuesday)
        weekDays = (6, 7, 1, 2, 3, 4, 5)
        return weekDays[date.weekday()]

    elif int(return_type) == 14:
        # Numbers 1 (Thursday) through 7 (Wednesday)
        weekDays = (5, 6, 7, 1, 2, 3, 4)
        return weekDays[date.weekday()]

    elif int(return_type) == 15:
        # Numbers 1 (Friday) through 7 (Thursday)
        weekDays = (4, 5, 6, 7, 1, 2, 3)
        return weekDays[date.weekday()]

    elif int(return_type) == 16:
        # Numbers 1 (Saturday) through 7 (Friday)
        weekDays = (3, 4, 5, 6, 7, 1, 2)
        return weekDays[date.weekday()]

    elif int(return_type) == 17:
        # Numbers 1 (Sunday) through 7 (Saturday)
        weekDays = (2, 3, 4, 5, 6, 7, 1)
        return weekDays[date.weekday()]

    else:
        raise xlerrors.NumExcelError(
            f"return_type needs to be omitted or one of 1, 2, 3, 11, 12, 13,\
            14, 15, 16 or 17. You supplied {return_type}")


@xl.register()
@xl.validate_args
def YEAR(
        serial_number: func_xltypes.XlNumber
) -> func_xltypes.XlNumber:
    """Returns the year corresponding to a date. The year is returned as an
        integer in the range 1900-9999.

    https://support.microsoft.com/en-us/office/
        year-function-c64f017a-1354-490d-981f-578e8ec8d3b9
    """

    date = utils.number_to_datetime(int(serial_number))

    if (int(date.strftime("%Y")) < int(utils.EXCEL_EPOCH.strftime("%Y"))) \
            or (int(date.strftime("%Y")) > 9999):
        raise xlerrors.ValueExcelError(
            f'year {date.strftime("%Y")} must be after \
            {utils.EXCEL_EPOCH.strftime("%Y")} and before 9999')

    return int(date.strftime("%Y"))


@xl.register()
@xl.validate_args
def YEARFRAC(
        start_date: func_xltypes.XlDateTime,
        end_date: func_xltypes.XlDateTime,
        basis: func_xltypes.XlNumber = 0
) -> func_xltypes.XlNumber:
    """Returns the fraction of the year represented by the number of whole
    days between two dates.

    https://support.office.com/en-us/article/
        yearfrac-function-3844141e-c76d-4143-82b6-208454ddc6a8
    """
    if start_date < utils.EXCEL_EPOCH:
        raise xlerrors.ValueExcelError(
            f'start_date {start_date} must be after {utils.EXCEL_EPOCH}')

    if end_date < utils.EXCEL_EPOCH:
        raise xlerrors.ValueExcelError(
            f'start_date {start_date} must be after {utils.EXCEL_EPOCH}')

    # Switch dates if start_date > end_date
    if start_date > end_date:
        start_date, end_date = end_date, start_date

    # Get Python internal types.
    start_date, end_date = start_date.value, end_date.value

    if basis == 0:  # US 30/360
        return yearfrac.yearfrac(start_date, end_date, '30e360_matu')
    elif basis == 1:  # Actual/actual
        return yearfrac.yearfrac(start_date, end_date, 'act_afb')
    elif basis == 2:  # Actual/360
        return (end_date - start_date).days / 360
    elif basis == 3:  # Actual/365
        return (end_date - start_date).days / 365
    elif basis == 4:  # Eurobond 30/360
        return yearfrac.yearfrac(start_date, end_date, '30e360')

    raise xlerrors.ValueExcelError(
        f'basis must be 0, 1, 2, 3 or 4, got {basis}')



import datetime
import operator
import unittest

from xlcalculator.xlfunctions import utils, xlerrors, func_xltypes


class ExcelTypeTest(unittest.TestCase):

    class MyType(func_xltypes.ExcelType):
        native_types = (int, type(None))

    def test__Blank__(self):
        blank = self.MyType(1).__Blank__()
        self.assertIsInstance(blank, func_xltypes.ExcelType)
        self.assertEqual(blank.value, None)

    def test__number__(self):
        self.assertEqual(self.MyType(1).__number__(), 1.0)

    def test__datetime__(self):
        with self.assertRaises(NotImplementedError):
            self.MyType(1).__datetime__()

    def test__repr__(self):
        self.assertEqual(repr(self.MyType(1)), '<MyType 1>')


class AbstractExcelTypeTest:

    value_zero = func_xltypes.Number(0)

    value1 = func_xltypes.Number(1)
    native1 = 1
    num1 = 1
    value1_float = 1
    value1_int = 1
    value1_text = '1'
    value1_bool = True
    value1_dt = datetime.datetime(1900, 1, 1)

    value2 = func_xltypes.Number(2)
    native2 = 2
    num2 = 2
    text2 = func_xltypes.Text('2')
    bool2 = func_xltypes.Boolean(True)
    dt2 = func_xltypes.DateTime(datetime.datetime(2020, 1, 1, 12, 0, 0))
    dt2_float = 6523831.0

    math_result_type = int

    def test__add__(self):
        res = self.value1 + self.value2
        self.assertIsNot(res, self.value1)
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, self.math_result_type)
        self.assertEqual(res.value, self.num1 + self.num2)

    def test__add___with_native_addend(self):
        res = self.value1 + self.native2
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, self.math_result_type)
        self.assertEqual(res.value, self.num1 + self.num2)

    def test__add__with_Text(self):
        res = self.value1 + self.text2
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, self.math_result_type)
        self.assertEqual(res.value, self.num1 + self.num2)

    def test__add__with_unconvertible_Text(self):
        with self.assertRaises(xlerrors.ValueExcelError):
            self.value1 + func_xltypes.Text('data')

    def test__add__with_Boolean(self):
        res = self.value1 + self.bool2
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, self.math_result_type)
        self.assertEqual(res.value, self.num1 + 1)

    def test__add__with_DateTime(self):
        res = self.value1 + self.dt2
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, float)
        self.assertEqual(res.value, self.num1 + self.dt2_float)

    def test__sub__(self):
        res = self.value1 - self.value2
        self.assertIsNot(res, self.value1)
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, self.math_result_type)
        self.assertEqual(res.value, self.num1 - self.num2)

    def test__sub___with_native_subtrahend(self):
        res = self.value1 - self.native2
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, self.math_result_type)
        self.assertEqual(res.value, self.num1 - self.num2)

    def test__mul__(self):
        res = self.value1 * self.value2
        self.assertIsNot(res, self.value1)
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, self.math_result_type)
        self.assertEqual(res.value, self.num1 * self.num2)

    def test__mul___with_native_factor(self):
        res = self.value1 * self.num2
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, self.math_result_type)
        self.assertEqual(res.value, self.num1 * self.num2)

    def test__mul__with_DateTime(self):
        self.assertEqual(
            (self.value1 * self.dt2).value, self.num1 * self.dt2_float
        )

    def test__truediv__(self):
        res = self.value1 / self.value2
        self.assertIsNot(res, self.value1)
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, float)
        self.assertEqual(res.value, self.num1 / self.num2)

    def test__truediv___by_zero(self):
        with self.assertRaises(xlerrors.DivZeroExcelError):
            self.value1 / 0

    def test__truediv___with_native_divisor(self):
        res = self.value1 / self.native2
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, float)
        self.assertEqual(res.value, self.num1 / self.num2)

    def test__truediv__with_DateTime(self):
        self.assertEqual(
            (self.value1 / self.dt2).value, self.num1 / self.dt2_float
        )

    def test__pow__(self):
        res = self.value1 ** self.value2
        self.assertIsNot(res, self.value1)
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, self.math_result_type)
        self.assertEqual(res.value, self.num1 ** self.num2)

    def test__pow___with_native_exponent(self):
        res = self.value1 ** self.num2
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, self.math_result_type)
        self.assertEqual(res.value, self.num1 ** self.num2)

    def test__pow__with_DateTime(self):
        self.assertEqual(
            (self.value1 ** self.dt2).value, self.num1 ** self.dt2_float
        )

    def test__lt__(self):
        res = self.value1 < self.value2
        self.assertIsInstance(res, func_xltypes.Boolean)
        self.assertIsInstance(res.value, bool)
        self.assertEqual(res.value, True)

    def test__lt___with_Text(self):
        self.assertEqual((self.value1 < func_xltypes.Text('data')).value, True)

    def test__lt___with_Boolean(self):
        self.assertEqual((self.value_zero < self.bool2).value, True)
        self.assertEqual((self.value2 < self.bool2).value, True)

    def test__lt___with_DateTime(self):
        self.assertEqual((self.value1 < self.dt2).value, True)

    def test__le__(self):
        self.assertEqual((self.value1 <= self.value2).value, True)
        self.assertEqual((self.value1 <= self.value1).value, True)
        self.assertEqual((self.value2 <= self.value1).value, False)

    def test__eq__(self):
        self.assertEqual((self.value1 == self.value2).value, False)
        self.assertEqual((self.value1 == self.value1).value, True)

    def test__ne__(self):
        self.assertEqual((self.value1 != self.value2).value, True)
        self.assertEqual((self.value1 != self.value1).value, False)

    def test__gt__(self):
        self.assertEqual((self.value1 > self.value2).value, False)
        self.assertEqual((self.value2 > self.value1).value, True)

    def test__ge__(self):
        self.assertEqual((self.value2 >= self.value1).value, True)
        self.assertEqual((self.value1 >= self.value1).value, True)
        self.assertEqual((self.value1 >= self.value2).value, False)

    def test__and__(self):
        self.assertEqual((self.value1 & self.value2).value, True)
        self.assertEqual((self.value1 & self.value_zero).value, False)

    def test__or__(self):
        self.assertEqual((self.value_zero | self.value1).value, True)
        self.assertEqual((self.value_zero | self.value_zero).value, False)

    def test__int__(self):
        self.assertEqual(int(self.value1), self.value1_int)
        self.assertEqual(
            int(func_xltypes.Number(self.value1_float)), self.value1_int)

    def test__float__(self):
        self.assertEqual(
            float(self.value1), self.value1_float)
        self.assertEqual(
            float(func_xltypes.Number(self.value1_float)), self.value1_float)

    def test__str__(self):
        self.assertEqual(str(self.value1), self.value1_text)

    def test__bool__(self):
        self.assertEqual(bool(self.value1), True)
        self.assertEqual(bool(self.value_zero), False)

    def test__Number__(self):
        self.assertEqual(self.value1.__Number__().value, self.num1)

    def test__Text__(self):
        text = self.value1.__Text__()
        self.assertIsInstance(text, func_xltypes.Text)
        self.assertEqual(text.value, self.value1_text)

    def test__Boolean__(self):
        boolean = self.value1.__Boolean__()
        self.assertIsInstance(boolean, func_xltypes.Boolean)
        self.assertEqual(boolean.value, True)

    def test__DateTime__(self):
        dt = self.value1.__DateTime__()
        self.assertIsInstance(dt, func_xltypes.DateTime)
        self.assertEqual(dt.value, self.value1_dt)

    def test__hash__(self):
        self.assertEqual(hash(self.value1), hash(self.value1.value))


class NumberTest(AbstractExcelTypeTest, unittest.TestCase):

    def test_is_whole(self):
        self.assertEqual(func_xltypes.Number(1).is_whole, True)
        self.assertEqual(func_xltypes.Number(1.0).is_whole, False)

    def test_is_decimal(self):
        self.assertEqual(func_xltypes.Number(1).is_decimal, False)
        self.assertEqual(func_xltypes.Number(1.0).is_decimal, True)

    def test__neg__(self):
        res = -func_xltypes.Number(2.0)
        self.assertIsInstance(res.value, float)
        self.assertEqual(res.value, -2.0)

    def test__pos__(self):
        res = +func_xltypes.Number(-2.0)
        self.assertIsInstance(res.value, float)
        self.assertEqual(res.value, -2.0)

    def test__invert__(self):
        res = operator.invert(func_xltypes.Number(2.0))
        self.assertIsInstance(res.value, float)
        self.assertEqual(res.value, -2.0)

    def test__add__with_decimal_number(self):
        res = func_xltypes.Number(1) + func_xltypes.Number(2.0)
        self.assertIsInstance(res.value, float)
        self.assertEqual(res.value, 3.0)

    def test__add__with_native_float(self):
        res = func_xltypes.Number(1) + 2.0
        self.assertIsInstance(res.value, float)
        self.assertEqual(res.value, 3.0)

    def test__sub__with_decimal_number(self):
        res = func_xltypes.Number(2) - func_xltypes.Number(1.0)
        self.assertIsInstance(res.value, float)
        self.assertEqual(res.value, 1.0)

    def test__sub__with_native_float(self):
        res = func_xltypes.Number(2) - 1.0
        self.assertIsInstance(res.value, float)
        self.assertEqual(res.value, 1.0)

    def test__lt___with_decimal_number(self):
        self.assertEqual(
            (self.value1 < func_xltypes.Number(2.0)).value, True)

    def test__lt___with_native_int(self):
        self.assertEqual(
            (self.value1 < 2).value, True)

    def test__lt___with_native_float(self):
        self.assertEqual(
            (self.value1 < 2.0).value, True)

    def test__lt___with_Blank(self):
        self.assertEqual((
            func_xltypes.Number(-1) < func_xltypes.BLANK).value, True)
        self.assertEqual((
            self.value1 < func_xltypes.BLANK).value, False)

    def test__Blank__(self):
        blank = self.value1.__Blank__()
        self.assertIsInstance(blank, func_xltypes.Number)
        self.assertEqual(blank.value, 0)

    def test__repr__(self):
        self.assertEqual(repr(self.value1), '<Number 1>')


class TextTest(AbstractExcelTypeTest, unittest.TestCase):

    value_zero = func_xltypes.Text('')

    value1 = func_xltypes.Text('1')
    native1 = '1'
    num1 = 1
    value1_float = 1.0
    value1_int = 1
    value1_text = '1'
    value1_bool = True
    value1_dt = datetime.datetime(1900, 1, 1)

    value2 = func_xltypes.Text('2')
    native2 = '2'
    num2 = 2
    text2 = func_xltypes.Text('2')
    bool2 = func_xltypes.Boolean(True)

    def test__add__with_unconvertible_Text(self):
        with self.assertRaises(xlerrors.ValueExcelError):
            self.value1 + func_xltypes.Text('data')

    def test__lt___with_Boolean(self):
        self.assertEqual((self.value1 < self.bool2).value, True)

    def test__lt___with_DateTime(self):
        # Text is always greater for comparison and is not converted.
        self.assertEqual((self.value1 < self.dt2).value, False)

    def test__int___with_unvonvertable_data(self):
        with self.assertRaises(xlerrors.ValueExcelError):
            int(func_xltypes.Text('data'))

    def test__float___with_unvonvertable_data(self):
        with self.assertRaises(xlerrors.ValueExcelError):
            float(func_xltypes.Text('data'))

    def test__number__(self):
        self.assertEqual(func_xltypes.Text('1').__number__(), 1)
        self.assertEqual(func_xltypes.Text('1.0').__number__(), 1.0)
        self.assertEqual(func_xltypes.Text('true').__number__(), 1)
        self.assertEqual(func_xltypes.Text('1900-01-01').__number__(), 1)

    def test__number___with_unvonvertable_data(self):
        with self.assertRaises(xlerrors.ValueExcelError):
            func_xltypes.Text('data').__number__()

    def test__bool__(self):
        self.assertEqual(bool(func_xltypes.Text('true')), True)
        self.assertEqual(bool(func_xltypes.Text('false')), False)
        self.assertEqual(bool(func_xltypes.Text('data')), True)
        self.assertEqual(bool(func_xltypes.Text('')), False)

    def test__datetime__(self):
        self.assertEqual(
            func_xltypes.Text('1').__datetime__(), self.value1_dt)
        self.assertEqual(
            func_xltypes.Text('1900-01-01').__datetime__(), self.value1_dt)

    def test__Boolean__(self):
        with self.assertRaises(xlerrors.ValueExcelError):
            self.value1.__Boolean__()
        boolean = func_xltypes.Text('true').__Boolean__()
        self.assertIsInstance(boolean, func_xltypes.Boolean)
        self.assertEqual(boolean.value, True)
        self.assertEqual(func_xltypes.Text('TRUE').__Boolean__().value, True)
        self.assertEqual(func_xltypes.Text('True').__Boolean__().value, True)
        self.assertEqual(func_xltypes.Text('False').__Boolean__().value, False)
        with self.assertRaises(xlerrors.ValueExcelError):
            self.assertEqual(func_xltypes.Text('').__Boolean__().value, False)

    def test__Blank__(self):
        blank = self.value1.__Blank__()
        self.assertIsInstance(blank, func_xltypes.Text)
        self.assertEqual(blank.value, '')


class BooleanTest(AbstractExcelTypeTest, unittest.TestCase):

    value_zero = func_xltypes.Boolean(False)

    value1 = func_xltypes.Boolean(True)
    native1 = True
    num1 = 1
    value1_float = 1.0
    value1_int = 1
    value1_text = 'True'
    value1_bool = True
    value1_dt = datetime.datetime(1999, 12, 31)

    value2 = func_xltypes.Boolean(True)
    native2 = True
    num2 = 1
    text2 = func_xltypes.Text('True')
    bool2 = func_xltypes.Boolean(True)

    def test__lt__(self):
        res = self.value_zero < self.value1
        self.assertIsInstance(res, func_xltypes.Boolean)
        self.assertIsInstance(res.value, bool)
        self.assertEqual(res.value, True)

    def test__lt___with_Text(self):
        self.assertEqual(
            (self.value1 < func_xltypes.Text('data')).value,
            False
        )

    def test__lt___with_Boolean(self):
        self.assertEqual((self.value_zero < self.bool2).value, True)

    def test__lt___with_DateTime(self):
        self.assertEqual((self.value1 < self.dt2).value, False)

    def test__le__(self):
        self.assertEqual((self.value_zero <= self.value2).value, True)
        self.assertEqual((self.value_zero <= self.value_zero).value, True)
        self.assertEqual((self.value2 <= self.value_zero).value, False)

    def test__eq__(self):
        self.assertEqual((self.value_zero == self.value2).value, False)
        self.assertEqual((self.value_zero == self.value_zero).value, True)

    def test__ne__(self):
        self.assertEqual((self.value_zero != self.value2).value, True)
        self.assertEqual((self.value_zero != self.value_zero).value, False)

    def test__gt__(self):
        self.assertEqual((self.value_zero > self.value2).value, False)
        self.assertEqual((self.value2 > self.value_zero).value, True)

    def test__ge__(self):
        self.assertEqual((self.value2 >= self.value_zero).value, True)
        self.assertEqual((self.value_zero >= self.value_zero).value, True)
        self.assertEqual((self.value_zero >= self.value2).value, False)

    def test__and__(self):
        self.assertEqual((self.value2 & self.value2).value, True)
        self.assertEqual((self.value2 & self.value_zero).value, False)

    def test__or__(self):
        self.assertEqual((self.value_zero | self.value2).value, True)
        self.assertEqual((self.value_zero | self.value_zero).value, False)

    def test__bool__(self):
        self.assertEqual(bool(self.value2), True)
        self.assertEqual(bool(self.value_zero), False)

    def test__Blank__(self):
        blank = self.value1.__Blank__()
        self.assertIsInstance(blank, func_xltypes.Boolean)
        self.assertEqual(blank.value, False)


class DateTimeTest(AbstractExcelTypeTest, unittest.TestCase):

    value_zero = func_xltypes.DateTime(utils.EXCEL_EPOCH)

    value1 = func_xltypes.DateTime(datetime.datetime(1900, 1, 1))
    native1 = datetime.datetime(1900, 1, 1)
    num1 = 1
    value1_float = 1.0
    value1_int = 1
    value1_text = '1900-01-01 00:00:00'
    value1_bool = True
    value1_dt = datetime.datetime(1900, 1, 1)

    value2 = func_xltypes.DateTime(datetime.datetime(1900, 2, 1))
    native2 = datetime.datetime(1900, 2, 1)
    num2 = 32
    text2 = func_xltypes.Text('1900-02-01 00:00:00')
    bool2 = func_xltypes.Boolean(True)

    math_result_type = float

    def test__sub__(self):
        res = self.value1 - self.value2
        self.assertIsNot(res, self.value1)
        self.assertIsInstance(res, func_xltypes.Number)
        self.assertIsInstance(res.value, float)
        self.assertEqual(res.value, self.num1 - self.num2)

    def test__bool__(self):
        self.assertEqual(bool(self.value1), True)
        self.assertEqual(bool(self.value_zero), True)

    def test__and__(self):
        # Bool of datetime will always be true.
        self.assertEqual((self.value1 & self.value2).value, True)
        self.assertEqual((self.value1 & self.value_zero).value, True)

    def test__or__(self):
        # Bool of datetime will always be true.
        self.assertEqual((self.value_zero | self.value1).value, True)
        self.assertEqual((self.value_zero | self.value_zero).value, True)

    def test__Blank__(self):
        blank = self.value1.__Blank__()
        self.assertIsNone(blank)

    def test__repr__(self):
        self.assertEqual(repr(self.value1), '<DateTime 1900-01-01T00:00:00>')


class BlankTest(unittest.TestCase):

    def test__add__(self):
        self.assertEqual((func_xltypes.BLANK + 1).value, 1)

    def test__sub__(self):
        self.assertEqual((func_xltypes.BLANK - 1).value, -1)

    def test__mul__(self):
        self.assertEqual((func_xltypes.BLANK * 1).value, 0)

    def test__truediv__(self):
        self.assertEqual((func_xltypes.BLANK / 1).value, 0)
        self.assertEqual((1 / func_xltypes.BLANK).value, 0)

    def test_pw__(self):
        self.assertEqual((func_xltypes.BLANK ** 1).value, 0)

    def test__lt__(self):
        self.assertEqual((func_xltypes.BLANK < 1).value, True)

    def test__le__(self):
        self.assertEqual((func_xltypes.BLANK <= 1).value, True)
        self.assertEqual((func_xltypes.BLANK <= 0).value, True)

    def test__eq__(self):
        self.assertEqual((func_xltypes.BLANK == 1).value, False)
        self.assertEqual((func_xltypes.BLANK == 0).value, True)
        self.assertEqual((func_xltypes.BLANK == 'data').value, False)
        self.assertEqual((func_xltypes.BLANK == '').value, True)
        self.assertEqual((func_xltypes.BLANK == True).value, False)    # noqa
        self.assertEqual((func_xltypes.BLANK == False).value, True)  # noqa

    def test__ne__(self):
        self.assertEqual((func_xltypes.BLANK != 1).value, True)
        self.assertEqual((func_xltypes.BLANK != 0).value, False)
        self.assertEqual((func_xltypes.BLANK != 'data').value, True)
        self.assertEqual((func_xltypes.BLANK != '').value, False)
        self.assertEqual((func_xltypes.BLANK != True).value, True)  # noqa
        self.assertEqual((func_xltypes.BLANK != False).value, False)  # noqa

    def test__gt__(self):
        self.assertEqual((func_xltypes.BLANK > 1).value, False)

    def test__ge__(self):
        self.assertEqual((func_xltypes.BLANK >= 1).value, False)
        self.assertEqual((func_xltypes.BLANK >= 0).value, True)

    def test__and__(self):
        self.assertEqual((func_xltypes.BLANK & 1).value, True)
        self.assertEqual((func_xltypes.BLANK & 1).value, True)
        with self.assertRaises(xlerrors.ValueExcelError):
            func_xltypes.BLANK & func_xltypes.BLANK

    def test__or__(self):
        self.assertEqual((func_xltypes.BLANK | 1).value, True)
        self.assertEqual((func_xltypes.BLANK | 1).value, True)
        with self.assertRaises(xlerrors.ValueExcelError):
            func_xltypes.BLANK | func_xltypes.BLANK

    def test__int__(self):
        self.assertEqual(int(func_xltypes.BLANK), 0)

    def test__float__(self):
        self.assertEqual(float(func_xltypes.BLANK), 0.0)

    def test__str__(self):
        self.assertEqual(str(func_xltypes.BLANK), '')

    def test__bool__(self):
        self.assertEqual(bool(func_xltypes.BLANK), False)

    def test__number__(self):
        self.assertEqual(func_xltypes.BLANK.__number__(), 0.0)

    def test__repr__(self):
        self.assertEqual(repr(func_xltypes.BLANK), '<BLANK>')


class ArrayTest(unittest.TestCase):

    def test__init__withbad_data(self):
        with self.assertRaises(xlerrors.ValueExcelError):
            func_xltypes.Array('bad')

    def test_flat(self):
        self.assertEqual(
            func_xltypes.Array([[1, 2, 3], [4, 5, 6]]).flat,
            [1, 2, 3, 4, 5, 6]
        )

    def test_cast_to_numbers(self):
        dt = datetime.datetime(1900, 1, 5)
        array = func_xltypes.Array([[1, None, dt], ['4.0', True, 6]])
        nums = array.cast_to_numbers()
        self.assertIsInstance(nums[0][0], func_xltypes.Number)
        self.assertEqual(nums.flat, [1, 0.0, 5.0, 4.0, 1, 6])

    def test_cast_to_numbers_with_errors(self):
        array = func_xltypes.Array([[1, object()]])
        nums = array.cast_to_numbers()
        self.assertEqual(nums.flat, [1, 0.0])

    def test_cast_to_booleans(self):
        dt = datetime.datetime(1900, 1, 5)
        array = func_xltypes.Array([[1, None, dt], ['4.0', True, 6]])
        bools = array.cast_to_booleans()
        self.assertIsInstance(bools[0][0], func_xltypes.Boolean)
        self.assertEqual(bools.flat, [True, False, True, True, True, True])

    def test_cast_to_texts(self):
        dt = datetime.datetime(1900, 1, 5)
        array = func_xltypes.Array([[1, None, dt], ['4.0', True, 6]])
        texts = array.cast_to_texts()
        self.assertIsInstance(texts[0][0], func_xltypes.Text)
        self.assertEqual(
            texts.flat,
            ['1', '', '1900-01-05 00:00:00', '4.0', 'True', '6'])


class ExprTest(unittest.TestCase):

    def test_init(self):

        def callable(arg):
            return arg

        expr = func_xltypes.Expr(callable, (1,), value=1)
        self.assertEqual(expr.callable, callable)
        self.assertEqual(expr.value, 1)

    def test_call(self):

        def callable(arg):
            return arg

        expr = func_xltypes.Expr(callable, (1,), value=1)
        self.assertEqual(expr(), 1)

    def test_ValueExpr(self):
        expr = func_xltypes.ValueExpr(1)
        self.assertEqual(expr.value, 1)
        self.assertEqual(expr(), 1)



from . import xl, xlerrors, func_xltypes


@xl.register()
@xl.validate_args
def OP_MUL(
        left: func_xltypes.XlAnything,
        right: func_xltypes.XlAnything
) -> func_xltypes.XlNumber:
    return left * right


@xl.register()
@xl.validate_args
def OP_DIV(
        left: func_xltypes.XlAnything,
        right: func_xltypes.XlAnything
) -> func_xltypes.XlNumber:
    if right == 0:
        raise xlerrors.DivZeroExcelError()
    return left / right


@xl.register()
@xl.validate_args
def OP_ADD(
        left: func_xltypes.XlAnything,
        right: func_xltypes.XlAnything
) -> func_xltypes.XlNumber:
    return left + right


@xl.register()
@xl.validate_args
def OP_SUB(
        left: func_xltypes.XlAnything,
        right: func_xltypes.XlAnything
) -> func_xltypes.XlNumber:
    return left - right


@xl.register()
def OP_EQ(
        left: func_xltypes.XlAnything,
        right: func_xltypes.XlAnything
) -> func_xltypes.XlBoolean:
    return left == right


@xl.register()
def OP_NE(
        left: func_xltypes.XlAnything,
        right: func_xltypes.XlAnything
) -> func_xltypes.XlBoolean:
    return left != right


@xl.register()
@xl.validate_args
def OP_GT(
        left: func_xltypes.XlAnything,
        right: func_xltypes.XlAnything
) -> func_xltypes.XlBoolean:
    if isinstance(left, func_xltypes.Blank) or \
            isinstance(right, func_xltypes.Blank):
        return False
    return left > right


@xl.register()
@xl.validate_args
def OP_LT(
        left: func_xltypes.XlAnything,
        right: func_xltypes.XlAnything
) -> func_xltypes.XlBoolean:
    if isinstance(left, func_xltypes.Blank) or \
            isinstance(right, func_xltypes.Blank):
        return False
    return left < right


@xl.register()
@xl.validate_args
def OP_GE(
        left: func_xltypes.XlAnything,
        right: func_xltypes.XlAnything
) -> func_xltypes.XlBoolean:
    if isinstance(left, func_xltypes.Blank) or \
            isinstance(right, func_xltypes.Blank):
        return False
    return left >= right


@xl.register()
@xl.validate_args
def OP_LE(
        left: func_xltypes.XlAnything,
        right: func_xltypes.XlAnything
) -> func_xltypes.XlBoolean:
    if isinstance(left, func_xltypes.Blank) or \
            isinstance(right, func_xltypes.Blank):
        return False
    return left <= right


@xl.register()
@xl.validate_args
def OP_NEG(
        right: func_xltypes.XlNumber
) -> func_xltypes.XlNumber:
    return -1 * right


@xl.register()
@xl.validate_args
def OP_PERCENT(
        left: func_xltypes.XlNumber
) -> func_xltypes.XlNumber:
    return left * 0.01
